from __future__ import annotations

from src.tasks.task import Task
from src.utils.exceptions import NotSupportedError
from src.wrappers.wrapper import Wrapper


class Version20(Wrapper):
    supported_versions: list[str] = ["2.1.0"]

    def __init__(self, task: Task, seed: int) -> None:
        super().__init__(task, seed)

        from smac.facade import AbstractFacade

        self._smac: AbstractFacade | None = None

    def run(self) -> None:
        from smac import (
            BlackBoxFacade,
            HyperparameterOptimizationFacade,
            MultiFidelityFacade,
            Scenario,
            AlgorithmConfigurationFacade
        )

        # Get instances
        instances = None
        instance_features = None

        if self.task.use_instances:
            instances = self.task.model.dataset.get_instances()
            instance_features = self.task.model.dataset.get_instance_features()

        # Create scenario
        scenario = Scenario(
            self.model.configspace,
            n_trials=self.task.n_trials,
            walltime_limit=self.task.walltime_limit,
            deterministic=self.task.deterministic,
            instances=instances,
            instance_features=instance_features,
            min_budget=self.task.min_budget,
            max_budget=self.task.max_budget,
            n_workers=self.task.n_workers,
            seed=self.seed,
        )

        intensifier_kwargs = {}

        # Create facade
        if self.task.optimization_type == "bb":
            facade_object = BlackBoxFacade
            intensifier_kwargs["max_config_calls"] = self.task.max_config_calls
        elif self.task.optimization_type == "hpo":
            facade_object = HyperparameterOptimizationFacade
            intensifier_kwargs["max_config_calls"] = self.task.max_config_calls
        elif self.task.optimization_type == "mf":
            facade_object = MultiFidelityFacade
            intensifier_kwargs["n_seeds"] = self.task.n_seeds
            intensifier_kwargs["incumbent_selection"] = self.task.incumbent_selection
        elif self.task.optimization_type == "ac":
            facade_object = AlgorithmConfigurationFacade
            intensifier_kwargs["max_config_calls"] = self.task.max_config_calls
        else:
            raise RuntimeError("Unknown optimization type.")

        if self.task.intensifier is None:
            intensifier = facade_object.get_intensifier(scenario, **intensifier_kwargs)
        else:
            if self.task.intensifier == "successive_halving":
                from smac.intensifier.successive_halving import SuccessiveHalving

                intensifier = SuccessiveHalving(scenario, **intensifier_kwargs)
            else:
                raise RuntimeError("Unsupported intensifier.")

        config_selector = facade_object.get_config_selector(scenario, retrain_after=self.task.retrain_after)

        smac = facade_object(
            scenario,
            self.task.model.train,
            intensifier=intensifier,
            config_selector=config_selector,
            logging_level=99999,
            overwrite=True,
        )
        smac.optimize()

        self._smac = smac

    def get_trajectory(self, sort_by: str = "trials") -> tuple[list[float], list[float]]:
        if len(self.task.objectives) > 1:
            raise NotSupportedError

        assert self._smac is not None
        rh = self._smac.runhistory
        trajectory = self._smac.intensifier.trajectory
        X: list[int | float] = []
        Y: list[float] = []

        for traj in trajectory:
            assert len(traj.config_ids) == 1
            config_id = traj.config_ids[0]
            config = rh.get_config(config_id)

            cost = rh.get_cost(config)
            if cost > 1e6:
                continue

            if sort_by == "trials":
                X.append(traj.trial)
            elif sort_by == "walltime":
                X.append(traj.walltime)
            else:
                raise RuntimeError("Unknown sort_by.")

            Y.append(cost)

        return X, Y



from __future__ import annotations

from ConfigSpace.configuration_space import Configuration
from ConfigSpace.hyperparameters import Constant
from scipy.stats.qmc import LatinHypercube

from smac.initial_design.abstract_initial_design import AbstractInitialDesign
from smac.utils.configspace import transform_continuous_designs

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class LatinHypercubeInitialDesign(AbstractInitialDesign):
    """Latin Hypercube initial design. See
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.qmc.LatinHypercube.html for further information.
    """

    def _select_configurations(self) -> list[Configuration]:
        params = list(self._configspace.values())

        constants = 0
        for p in params:
            if isinstance(p, Constant):
                constants += 1

        lhd = LatinHypercube(d=len(params) - constants, seed=self._rng.randint(0, 1000000)).random(n=self._n_configs)

        return transform_continuous_designs(
            design=lhd, origin="Initial Design: Latin Hypercube", configspace=self._configspace
        )



from __future__ import annotations

from typing import Any

import numpy as np
from ConfigSpace import (
    CategoricalHyperparameter,
    Constant,
    OrdinalHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
)

from smac.model.abstract_model import AbstractModel

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class AbstractRandomForest(AbstractModel):
    """Abstract base class for all random forest models."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self._conditional: dict[int, bool] = dict()
        self._impute_values: dict[int, float] = dict()

    def _impute_inactive(self, X: np.ndarray) -> np.ndarray:
        X = X.copy()
        for idx, hp in enumerate(list(self._configspace.values())):
            if idx not in self._conditional:
                parents = self._configspace.parents_of[hp.name]
                if len(parents) == 0:
                    self._conditional[idx] = False
                else:
                    self._conditional[idx] = True
                    if isinstance(hp, CategoricalHyperparameter):
                        self._impute_values[idx] = len(hp.choices)
                    elif isinstance(hp, OrdinalHyperparameter):
                        self._impute_values[idx] = len(hp.sequence)
                    elif isinstance(hp, (UniformFloatHyperparameter, UniformIntegerHyperparameter)):
                        self._impute_values[idx] = -1
                    elif isinstance(hp, Constant):
                        self._impute_values[idx] = 1
                    else:
                        raise ValueError(f"Unsupported hyperparameter type: {type(hp)}")

            if self._conditional[idx] is True:
                nonfinite_mask = ~np.isfinite(X[:, idx])
                X[nonfinite_mask, idx] = self._impute_values[idx]

        return X



from __future__ import annotations

from abc import abstractmethod
from typing import Any, TypeVar

import copy
import warnings

import numpy as np
from ConfigSpace import ConfigurationSpace
from sklearn.decomposition import PCA
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import MinMaxScaler

from smac.constants import VERY_SMALL_NUMBER
from smac.utils.configspace import get_types
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


logger = get_logger(__name__)


Self = TypeVar("Self", bound="AbstractModel")


class AbstractModel:
    """Abstract implementation of the surrogate model.

    Note
    ----
    The input dimensionality of Y for training and the output dimensions of all predictions depend on the concrete
    implementation of this abstract class.

    Parameters
    ----------
    configspace : ConfigurationSpace
    instance_features : dict[str, list[int | float]] | None, defaults to None
        Features (list of int or floats) of the instances (str). The features are incorporated into the X data,
        on which the model is trained on.
    pca_components : float, defaults to 7
        Number of components to keep when using PCA to reduce dimensionality of instance features.
    seed : int
    """

    def __init__(
        self,
        configspace: ConfigurationSpace,
        instance_features: dict[str, list[int | float]] | None = None,
        pca_components: int | None = 7,
        seed: int = 0,
    ) -> None:
        self._configspace = configspace
        self._seed = seed
        self._rng = np.random.RandomState(self._seed)
        self._instance_features = instance_features
        self._pca_components = pca_components

        n_features = 0
        if self._instance_features is not None:
            for v in self._instance_features.values():
                if n_features == 0:
                    n_features = len(v)
                else:
                    if len(v) != n_features:
                        raise RuntimeError("Instances must have the same number of features.")

        self._n_features = n_features
        self._n_hps = len(list(self._configspace.values()))

        self._pca = PCA(n_components=self._pca_components)
        self._scaler = MinMaxScaler()
        self._apply_pca = False

        # Never use a lower variance than this.
        # If estimated variance < var_threshold, set to var_threshold
        self._var_threshold = VERY_SMALL_NUMBER
        self._types, self._bounds = get_types(configspace, instance_features)

        # Initial types array which is used to reset the type array at every call to `self.train()`
        self._initial_types = copy.deepcopy(self._types)

    @property
    def meta(self) -> dict[str, Any]:
        """Returns the meta data of the created object."""
        return {
            "name": self.__class__.__name__,
            "types": self._types,
            "bounds": self._bounds,
            "pca_components": self._pca_components,
        }

    def train(self: Self, X: np.ndarray, Y: np.ndarray) -> Self:
        """Trains the random forest on X and Y. Internally, calls the method `_train`.

        Parameters
        ----------
        X : np.ndarray [#samples, #hyperparameters + #features]
            Input data points.
        Y : np.ndarray [#samples, #objectives]
            The corresponding target values.

        Returns
        -------
        self : AbstractModel
        """
        if len(X.shape) != 2:
            raise ValueError("Expected 2d array, got %dd array!" % len(X.shape))

        if X.shape[1] != self._n_hps + self._n_features:
            raise ValueError(
                f"Feature mismatch: X should have {self._n_hps} hyperparameters + {self._n_features} features, "
                f"but has {X.shape[1]} in total."
            )

        if X.shape[0] != Y.shape[0]:
            raise ValueError("X.shape[0] ({}) != y.shape[0] ({})".format(X.shape[0], Y.shape[0]))

        # Reduce dimensionality of features if larger than PCA_DIM
        if (
            self._pca_components is not None
            and X.shape[0] > self._pca.n_components
            and self._n_features >= self._pca_components
        ):
            X_feats = X[:, -self._n_features :]

            # Scale features
            X_feats = self._scaler.fit_transform(X_feats)
            X_feats = np.nan_to_num(X_feats)  # if features with max == min

            # PCA
            X_feats = self._pca.fit_transform(X_feats)
            X = np.hstack((X[:, : self._n_hps], X_feats))

            if hasattr(self, "_types"):
                # For RF, adapt types list
                # if X_feats.shape[0] < self._pca, X_feats.shape[1] == X_feats.shape[0]
                self._types = np.array(
                    np.hstack((self._types[: self._n_hps], np.zeros(X_feats.shape[1]))),
                    dtype=np.uint,
                )  # type: ignore

            self._apply_pca = True
        else:
            self._apply_pca = False

            if hasattr(self, "_types"):
                self._types = copy.deepcopy(self._initial_types)

        return self._train(X, Y)

    @abstractmethod
    def _train(self: Self, X: np.ndarray, Y: np.ndarray) -> Self:
        """Trains the random forest on X and Y.

        Parameters
        ----------
        X : np.ndarray [#samples, #hyperparameters + #features]
            Input data points.
        Y : np.ndarray [#samples, #objectives]
            The corresponding target values.

        Returns
        -------
        self : AbstractModel
        """
        raise NotImplementedError()

    def predict(
        self,
        X: np.ndarray,
        covariance_type: str | None = "diagonal",
    ) -> tuple[np.ndarray, np.ndarray | None]:
        """Predicts mean and variance for a given X. Internally, calls the method `_predict`.

        Parameters
        ----------
        X : np.ndarray [#samples, #hyperparameters + #features]
            Input data points.
        covariance_type: str | None, defaults to "diagonal"
            Specifies what to return along with the mean. Applied only to Gaussian Processes.
            Takes four valid inputs:
            * None: Only the mean is returned.
            * "std": Standard deviation at test points is returned.
            * "diagonal": Diagonal of the covariance matrix is returned.
            * "full": Whole covariance matrix between the test points is returned.

        Returns
        -------
        means : np.ndarray [#samples, #objectives]
            The predictive mean.
        vars : np.ndarray [#samples, #objectives] or [#samples, #samples] | None
            Predictive variance or standard deviation.
        """
        if len(X.shape) != 2:
            raise ValueError("Expected 2d array, got %dd array!" % len(X.shape))

        if X.shape[1] != self._n_hps + self._n_features:
            raise ValueError(
                f"Feature mismatch: X should have {self._n_hps} hyperparameters + {self._n_features} features, "
                f"but has {X.shape[1]} in total."
            )

        if self._apply_pca:
            try:
                X_feats = X[:, -self._n_features :]
                X_feats = self._scaler.transform(X_feats)
                X_feats = self._pca.transform(X_feats)
                X = np.hstack((X[:, : self._n_hps], X_feats))
            except NotFittedError:
                # PCA not fitted if only one training sample
                pass

        if X.shape[1] != len(self._types):
            raise ValueError("Rows in X should have %d entries but have %d!" % (len(self._types), X.shape[1]))

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "Predicted variances smaller than 0. Setting those variances to 0.")
            mean, var = self._predict(X, covariance_type)

        if len(mean.shape) == 1:
            mean = mean.reshape((-1, 1))

        if var is not None and len(var.shape) == 1:
            var = var.reshape((-1, 1))

        return mean, var

    def _predict(
        self,
        X: np.ndarray,
        covariance_type: str | None = "diagonal",
    ) -> tuple[np.ndarray, np.ndarray | None]:
        """Predicts mean and variance for a given X.

        Parameters
        ----------
        X : np.ndarray [#samples, #hyperparameters + #features]
            Input data points.
        covariance_type : str | None, defaults to "diagonal"
            Specifies what to return along with the mean. Applied only to Gaussian Processes.
            Takes four valid inputs:
            * None: Only the mean is returned.
            * "std": Standard deviation at test points is returned.
            * "diagonal": Diagonal of the covariance matrix is returned.
            * "full": Whole covariance matrix between the test points is returned.

        Returns
        -------
        means : np.ndarray [#samples, #objectives]
            The predictive mean.
        vars : np.ndarray [#samples, #objectives] or [#samples, #samples] | None
            Predictive variance or standard deviation.
        """
        raise NotImplementedError()

    def predict_marginalized(self, X: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        """Predicts mean and variance marginalized over all instances.

        Warning
        -------
        The input data must not include any features.

        Parameters
        ----------
        X : np.ndarray [#samples, #hyperparameters]
            Input data points.

        Returns
        -------
        means : np.ndarray [#samples, 1]
            The predictive mean.
        vars : np.ndarray [#samples, 1]
            The predictive variance.
        """
        if len(X.shape) != 2:
            raise ValueError("Expected 2d array, got %dd array!" % len(X.shape))

        if X.shape[1] != self._n_hps:
            raise ValueError(
                f"Feature mismatch: X should have {self._n_hps} hyperparameters (and no features) for this method, "
                f"but has {X.shape[1]} in total."
            )

        if self._instance_features is None:
            mean, var = self.predict(X)
            assert var is not None

            var[var < self._var_threshold] = self._var_threshold
            var[np.isnan(var)] = self._var_threshold

            return mean, var
        else:
            n_instances = len(self._instance_features)

            mean = np.zeros(X.shape[0])
            var = np.zeros(X.shape[0])
            for i, x in enumerate(X):
                features = np.array(list(self._instance_features.values()))
                x_tiled = np.tile(x, (n_instances, 1))
                X_ = np.hstack((x_tiled, features))

                means, vars = self.predict(X_)
                assert vars is not None

                # VAR[1/n (X_1 + ... + X_n)] =
                # 1/n^2 * ( VAR(X_1) + ... + VAR(X_n))
                # for independent X_1 ... X_n
                var_x = np.sum(vars) / (len(vars) ** 2)
                if var_x < self._var_threshold:
                    var_x = self._var_threshold

                var[i] = var_x
                mean[i] = np.mean(means)

            if len(mean.shape) == 1:
                mean = mean.reshape((-1, 1))

            if len(var.shape) == 1:
                var = var.reshape((-1, 1))

            return mean, var



from smac.model.gaussian_process.abstract_gaussian_process import (
    AbstractGaussianProcess,
)
from smac.model.gaussian_process.gaussian_process import GaussianProcess
from smac.model.gaussian_process.mcmc_gaussian_process import MCMCGaussianProcess

__all__ = [
    "AbstractGaussianProcess",
    "GaussianProcess",
    "MCMCGaussianProcess",
]



from __future__ import annotations

from abc import abstractmethod
from typing import Any

import numpy as np
import sklearn.gaussian_process
from ConfigSpace import ConfigurationSpace
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Kernel, KernelOperator

from smac.model.abstract_model import AbstractModel
from smac.model.gaussian_process.priors.abstract_prior import AbstractPrior
from smac.model.gaussian_process.priors.tophat_prior import SoftTopHatPrior, TophatPrior

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class AbstractGaussianProcess(AbstractModel):
    """Abstract base class for all Gaussian process models.

    Parameters
    ----------
    configspace : ConfigurationSpace
    kernel : Kernel
        Kernel which is used for the Gaussian process.
    instance_features : dict[str, list[int | float]] | None, defaults to None
        Features (list of int or floats) of the instances (str). The features are incorporated into the X data,
        on which the model is trained on.
    pca_components : float, defaults to 7
        Number of components to keep when using PCA to reduce dimensionality of instance features.
    seed : int
    """

    def __init__(
        self,
        configspace: ConfigurationSpace,
        kernel: Kernel,
        instance_features: dict[str, list[int | float]] | None = None,
        pca_components: int | None = 7,
        seed: int = 0,
    ):
        super().__init__(
            configspace=configspace,
            instance_features=instance_features,
            pca_components=pca_components,
            seed=seed,
        )

        self._kernel = kernel
        self._gp = self._get_gaussian_process()

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update({"kernel": self._kernel.meta})

        return meta

    @abstractmethod
    def _get_gaussian_process(self) -> GaussianProcessRegressor:
        """Generates a Gaussian process."""
        raise NotImplementedError()

    def _normalize(self, y: np.ndarray) -> np.ndarray:
        """Normalize data to zero mean unit standard deviation.

        Parameters
        ----------
        y : np.ndarray
            Target values for the Gaussian process.

        Returns
        -------
        normalized_y : np.ndarray
            Normalized y values.
        """
        self.mean_y_ = np.mean(y)
        self.std_y_ = np.std(y)

        if self.std_y_ == 0:
            self.std_y_ = 1

        return (y - self.mean_y_) / self.std_y_

    def _untransform_y(
        self,
        y: np.ndarray,
        var: np.ndarray | None = None,
    ) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """Transform zero mean unit standard deviation data into the regular space.

        Warning
        -------
        This function should be used after a prediction with the Gaussian process which was
        trained on normalized data.

        Parameters
        ----------
        y : np.ndarray
            Normalized data.
        var : np.ndarray | None, defaults to None
            Normalized variance.

        Returns
        -------
        untransformed_y : np.ndarray | tuple[np.ndarray, np.ndarray]
        """
        y = y * self.std_y_ + self.mean_y_
        if var is not None:
            var = var * self.std_y_**2
            return y, var  # type: ignore

        return y

    def _get_all_priors(
        self,
        add_bound_priors: bool = True,
        add_soft_bounds: bool = False,
    ) -> list[list[AbstractPrior]]:
        """Returns all priors."""
        # Obtain a list of all priors for each tunable hyperparameter of the kernel
        all_priors = []
        to_visit = []
        to_visit.append(self._gp.kernel.k1)
        to_visit.append(self._gp.kernel.k2)

        while len(to_visit) > 0:
            current_param = to_visit.pop(0)
            if isinstance(current_param, KernelOperator):
                to_visit.insert(0, current_param.k1)
                to_visit.insert(1, current_param.k2)
                continue
            elif isinstance(current_param, Kernel):
                hps = current_param.hyperparameters
                assert len(hps) == 1

                hp = hps[0]
                if hp.fixed:
                    continue

                bounds = hps[0].bounds
                for i in range(hps[0].n_elements):
                    priors_for_hp = []

                    if current_param.prior is not None:
                        priors_for_hp.append(current_param.prior)

                    if add_bound_priors:
                        if add_soft_bounds:
                            priors_for_hp.append(
                                SoftTopHatPrior(
                                    lower_bound=bounds[i][0],
                                    upper_bound=bounds[i][1],
                                    seed=self._rng.randint(0, 2**20),
                                    exponent=2,
                                )
                            )
                        else:
                            priors_for_hp.append(
                                TophatPrior(
                                    lower_bound=bounds[i][0],
                                    upper_bound=bounds[i][1],
                                    seed=self._rng.randint(0, 2**20),
                                )
                            )
                    all_priors.append(priors_for_hp)

        return all_priors

    def _set_has_conditions(self) -> None:
        """Sets `has_conditions` on `current_param`."""
        has_conditions = len(self._configspace.conditions) > 0
        to_visit = []
        to_visit.append(self._kernel)

        while len(to_visit) > 0:
            current_param = to_visit.pop(0)
            if isinstance(current_param, sklearn.gaussian_process.kernels.KernelOperator):
                to_visit.insert(0, current_param.k1)
                to_visit.insert(1, current_param.k2)
                current_param.has_conditions = has_conditions
            elif isinstance(current_param, sklearn.gaussian_process.kernels.Kernel):
                current_param.has_conditions = has_conditions
            else:
                raise ValueError(current_param)

    def _impute_inactive(self, X: np.ndarray) -> np.ndarray:
        """Imputes inactives."""
        X = X.copy()
        X[~np.isfinite(X)] = -1

        return X



from __future__ import annotations

from abc import abstractmethod
from typing import Any

from collections import OrderedDict

import numpy as np
from ConfigSpace.configuration_space import Configuration

from smac.scenario import Scenario
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class AbstractInitialDesign:
    """Base class for initial design strategies that evaluates multiple configurations.

    Parameters
    ----------
    scenario : Scenario
    n_configs : int | None, defaults to None
        Number of initial configurations (disables the arguments ``n_configs_per_hyperparameter``).
    n_configs_per_hyperparameter: int, defaults to 10
        Number of initial configurations per hyperparameter. For example, if my configuration space covers five
        hyperparameters and ``n_configs_per_hyperparameter`` is set to 10, then 50 initial configurations will be
        samples.
    max_ratio: float, defaults to 0.25
        Use at most ``scenario.n_trials`` * ``max_ratio`` number of configurations in the initial design.
        Additional configurations are not affected by this parameter.
    additional_configs: list[Configuration], defaults to []
        Adds additional configurations to the initial design.
    seed : int | None, default to None
    """

    def __init__(
        self,
        scenario: Scenario,
        n_configs: int | None = None,
        n_configs_per_hyperparameter: int | None = 10,
        max_ratio: float = 0.25,
        additional_configs: list[Configuration] = None,
        seed: int | None = None,
    ):
        self._configspace = scenario.configspace

        if seed is None:
            seed = scenario.seed

        self.use_default_config = scenario.use_default_config

        self._seed = seed
        self._rng = np.random.RandomState(seed)
        self._n_configs_per_hyperparameter = n_configs_per_hyperparameter

        # make sure that additional configs is not a mutable default value
        # this avoids issues
        if additional_configs is None:
            additional_configs = []

        if self.use_default_config:
            default_config = self._configspace.get_default_configuration()
            default_config.origin = "Initial Design: Default configuration"
            additional_configs.append(default_config)

        self._additional_configs = additional_configs

        n_params = len(list(self._configspace.values()))
        if n_configs is not None:
            logger.info("Using `n_configs` and ignoring `n_configs_per_hyperparameter`.")
            self._n_configs = n_configs
        elif n_configs_per_hyperparameter is not None:
            self._n_configs = n_configs_per_hyperparameter * n_params
        else:
            raise ValueError(
                "Need to provide either argument `n_configs` or "
                "`n_configs_per_hyperparameter` but provided none of them."
            )

        # If the number of configurations is too large, we reduce it
        _n_configs = int(max(1, min(self._n_configs, (max_ratio * scenario.n_trials))))
        if self._n_configs != _n_configs:
            logger.info(
                f"Reducing the number of initial configurations from {self._n_configs} to "
                f"{_n_configs} (max_ratio == {max_ratio})."
            )
            self._n_configs = _n_configs

        # We allow no configs if we have additional configs
        if n_configs is not None and n_configs == 0 and len(additional_configs) > 0:
            self._n_configs = 0

        if self._n_configs + len(additional_configs) > scenario.n_trials:
            raise ValueError(
                f"Initial budget {self._n_configs} cannot be higher than the number of trials {scenario.n_trials}."
            )

    @property
    def meta(self) -> dict[str, Any]:
        """Returns the meta data of the created object."""
        return {
            "name": self.__class__.__name__,
            "n_configs": self._n_configs,
            "n_configs_per_hyperparameter": self._n_configs_per_hyperparameter,
            "additional_configs": [c.get_dictionary() for c in self._additional_configs],
            "seed": self._seed,
        }

    def select_configurations(self) -> list[Configuration]:
        """Selects the initial configurations. Internally, `_select_configurations` is called,
        which has to be implemented by the child class.

        Returns
        -------
        configs : list[Configuration]
            Configurations from the child class.
        """
        configs: list[Configuration] = []

        if self._n_configs == 0:
            logger.info("No initial configurations are used.")
        else:
            configs += self._select_configurations()

        # Adding additional configs
        configs += self._additional_configs

        for config in configs:
            if config.origin is None:
                config.origin = "Initial design"

        # Removing duplicates
        # (Reference: https://stackoverflow.com/questions/7961363/removing-duplicates-in-lists)
        configs = list(OrderedDict.fromkeys(configs))
        logger.info(
            f"Using {len(configs) - len(self._additional_configs)} initial design configurations "
            f"and {len(self._additional_configs)} additional configurations."
        )

        return configs

    @abstractmethod
    def _select_configurations(self) -> list[Configuration]:
        """Selects the initial configurations, depending on the implementation of the initial design."""
        raise NotImplementedError



from __future__ import annotations

import numpy as np
from ConfigSpace import Configuration

from smac.runhistory import RunHistory
from smac.runhistory.dataclasses import InstanceSeedBudgetKey


def _get_costs(
    runhistory: RunHistory,
    configs: list[Configuration],
    config_instance_seed_budget_keys: list[list[InstanceSeedBudgetKey]],
) -> np.ndarray:
    """Returns the costs of the passed configurations.

    Parameters
    ----------
    runhistory : RunHistory
        The runhistory containing the passed configs.
    configs : list[Configuration]
        The configs for which the costs should be returned.
    config_instance_seed_budget_keys: list[list[InstanceSeedBudgetKey]]
        The instance-seed budget keys for the configs for which the costs should be returned.

    Returns
    -------
    costs : np.ndarray[n_points, n_objectives]
        Costs of the given configs.
    """
    assert len(configs) == len(config_instance_seed_budget_keys)

    # Now we get the costs for the trials of the config
    average_costs = []

    for config, isb_keys in zip(configs, config_instance_seed_budget_keys):
        # Since we use multiple seeds, we have to average them to get only one cost value pair for each
        # configuration
        # However, we only want to consider the config trials
        # Average cost is a list of floats (one for each objective)
        average_cost = runhistory.average_cost(config, isb_keys, normalize=False)
        average_costs += [average_cost]

    # Let's work with a numpy array for efficiency
    return np.vstack(average_costs)


def calculate_pareto_front(
    runhistory: RunHistory,
    configs: list[Configuration],
    config_instance_seed_budget_keys: list[list[InstanceSeedBudgetKey]],
) -> list[Configuration]:
    """Compares the passed configurations and returns only the ones on the pareto front.

    Parameters
    ----------
    runhistory : RunHistory
        The runhistory containing the given configurations.
    configs : list[Configuration]
        The configurations from which the Pareto front should be computed.
    config_instance_seed_budget_keys: list[list[InstanceSeedBudgetKey]]
        The instance-seed budget keys for the configurations on the basis of which the Pareto front should be computed.

    Returns
    -------
    pareto_front : list[Configuration]
        The pareto front computed from the given configurations.
    """
    costs = _get_costs(runhistory, configs, config_instance_seed_budget_keys)

    # The following code is an efficient pareto front implementation
    is_efficient = np.arange(costs.shape[0])
    next_point_index = 0  # Next index in the is_efficient array to search for
    while next_point_index < len(costs):
        nondominated_point_mask = np.any(costs < costs[next_point_index], axis=1)
        nondominated_point_mask[next_point_index] = True
        is_efficient = is_efficient[nondominated_point_mask]  # Remove dominated points
        costs = costs[nondominated_point_mask]
        next_point_index = np.sum(nondominated_point_mask[:next_point_index]) + 1

    new_incumbents = [configs[i] for i in is_efficient]
    return new_incumbents


def sort_by_crowding_distance(
    runhistory: RunHistory,
    configs: list[Configuration],
    config_instance_seed_budget_keys: list[list[InstanceSeedBudgetKey]],
) -> list[Configuration]:
    """Sorts the passed configurations by their crowding distance. Taken from
    https://github.com/anyoptimization/pymoo/blob/20abef1ade71915352217400c11ece4c2f35163e/pymoo/algorithms/nsga2.py


    Parameters
    ----------
    runhistory : RunHistory
        The runhistory containing the given configurations.
    configs : list[Configuration]
        The configurations which should be sorted.
    config_instance_seed_budget_keys: list[list[InstanceSeedBudgetKey]]
        The instance-seed budget keys for the configurations which should be sorted.

    Returns
    -------
    sorted_list : list[Configuration]
        Configurations sorted by crowding distance.
    """
    F = _get_costs(runhistory, configs, config_instance_seed_budget_keys)
    infinity = 1e14

    n_points = F.shape[0]
    n_obj = F.shape[1]

    if n_points <= 2:
        # distances = np.full(n_points, infinity)
        return configs
    else:
        # Sort each column and get index
        I = np.argsort(F, axis=0, kind="mergesort")  # noqa

        # Now really sort the whole array
        F = F[I, np.arange(n_obj)]

        # get the distance to the last element in sorted list and replace zeros with actual values
        dist = np.concatenate([F, np.full((1, n_obj), np.inf)]) - np.concatenate([np.full((1, n_obj), -np.inf), F])

        index_dist_is_zero = np.where(dist == 0)

        dist_to_last = np.copy(dist)
        for i, j in zip(*index_dist_is_zero):
            dist_to_last[i, j] = dist_to_last[i - 1, j]

        dist_to_next = np.copy(dist)
        for i, j in reversed(list(zip(*index_dist_is_zero))):
            dist_to_next[i, j] = dist_to_next[i + 1, j]

        # Normalize all the distances
        norm = np.max(F, axis=0) - np.min(F, axis=0)
        norm[norm == 0] = np.nan
        dist_to_last, dist_to_next = dist_to_last[:-1] / norm, dist_to_next[1:] / norm

        # If we divided by zero because all values in one columns are equal replace by none
        dist_to_last[np.isnan(dist_to_last)] = 0.0
        dist_to_next[np.isnan(dist_to_next)] = 0.0

        # Sum up the distance to next and last and norm by objectives - also reorder from sorted list
        J = np.argsort(I, axis=0)
        crowding = np.sum(dist_to_last[J, np.arange(n_obj)] + dist_to_next[J, np.arange(n_obj)], axis=1) / n_obj

    # Replace infinity with a large number
    crowding[np.isinf(crowding)] = infinity
    config_with_crowding = [(config, v) for config, v in zip(configs, crowding)]
    config_with_crowding = sorted(config_with_crowding, key=lambda x: x[1], reverse=True)

    return [c for c, _ in config_with_crowding]



from __future__ import annotations

from typing import Any

import itertools
import time

import numpy as np
from ConfigSpace import Configuration, ConfigurationSpace
from ConfigSpace.exceptions import ForbiddenValueError

from smac.acquisition.function import AbstractAcquisitionFunction
from smac.acquisition.maximizer.abstract_acquisition_maximizer import (
    AbstractAcquisitionMaximizer,
)
from smac.utils.configspace import (
    convert_configurations_to_array,
    get_one_exchange_neighbourhood,
)
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class LocalSearch(AbstractAcquisitionMaximizer):
    """Implementation of SMAC's local search.

    Parameters
    ----------
    configspace : ConfigurationSpace
    acquisition_function : AbstractAcquisitionFunction
    challengers : int, defaults to 5000
        Number of challengers.
    max_steps: int | None, defaults to None
        Maximum number of iterations that the local search will perform.
    n_steps_plateau_walk: int, defaults to 10
        Number of steps during a plateau walk before local search terminates.
    vectorization_min_obtain : int, defaults to 2
        Minimal number of neighbors to obtain at once for each local search for vectorized calls. Can be tuned to
        reduce the overhead of SMAC.
    vectorization_max_obtain : int, defaults to 64
        Maximal number of neighbors to obtain at once for each local search for vectorized calls. Can be tuned to
        reduce the overhead of SMAC.
    seed : int, defaults to 0
    """

    def __init__(
        self,
        configspace: ConfigurationSpace,
        acquisition_function: AbstractAcquisitionFunction | None = None,
        challengers: int = 5000,
        max_steps: int | None = None,
        n_steps_plateau_walk: int = 10,
        vectorization_min_obtain: int = 2,
        vectorization_max_obtain: int = 64,
        seed: int = 0,
    ) -> None:
        super().__init__(
            configspace,
            acquisition_function,
            challengers=challengers,
            seed=seed,
        )

        self._max_steps = max_steps
        self._n_steps_plateau_walk = n_steps_plateau_walk
        self._vectorization_min_obtain = vectorization_min_obtain
        self._vectorization_max_obtain = vectorization_max_obtain

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update(
            {
                "max_steps": self._max_steps,
                "n_steps_plateau_walk": self._n_steps_plateau_walk,
                "vectorization_min_obtain": self._vectorization_min_obtain,
                "vectorization_max_obtain": self._vectorization_max_obtain,
            }
        )

        return meta

    def _maximize(
        self,
        previous_configs: list[Configuration],
        n_points: int,
        additional_start_points: list[tuple[float, Configuration]] | None = None,
    ) -> list[tuple[float, Configuration]]:
        """Start a local search from the given start points. Iteratively collect neighbours
        using Configspace.utils.get_one_exchange_neighbourhood and evaluate them.
        If the new config is better than the current best, the local search is continued from the
        new config.

        Quit if either the max number of steps is reached or
        no neighbor with a higher improvement was found or the number of local steps self._n_steps_plateau_walk
        for each of the starting point is depleted.


        Parameters
        ----------
        previous_configs : list[Configuration]
            Previous configuration (e.g., from the runhistory).
        n_points : int
            Number of initial points to be generated.
        additional_start_points : list[tuple[float, Configuration]] | None
            Additional starting points.

        Returns
        -------
        list[Configuration]
            Final candidates.
        """
        init_points = self._get_initial_points(previous_configs, n_points, additional_start_points)
        configs_acq = self._search(init_points)

        # Shuffle for random tie-break
        self._rng.shuffle(configs_acq)

        # Sort according to acq value
        configs_acq.sort(reverse=True, key=lambda x: x[0])
        for a, inc in configs_acq:
            inc.origin = "Acquisition Function Maximizer: Local Search"

        return configs_acq

    def _get_initial_points(
        self,
        previous_configs: list[Configuration],
        n_points: int,
        additional_start_points: list[tuple[float, Configuration]] | None,
    ) -> list[Configuration]:
        """Get initial points to start search from.

        Parameters
        ----------
        previous_configs : list[Configuration]
            Previous configuration (e.g., from the runhistory).
        n_points : int
            Number of initial points to be generated.
        additional_start_points : list[tuple[float, Configuration]] | None
            Additional starting points.

        Returns
        -------
        list[Configuration]
            A list of initial points/configurations.
        """
        sampled_points = []
        init_points = []
        n_init_points = n_points
        if len(previous_configs) < n_points:
            if n_points - len(previous_configs) == 1:
                sampled_points = [self._configspace.sample_configuration()]
            else:
                sampled_points = self._configspace.sample_configuration(size=n_points - len(previous_configs))
            n_init_points = len(previous_configs)
            if not isinstance(sampled_points, list):
                sampled_points = [sampled_points]
        if len(previous_configs) > 0:
            init_points = self._get_init_points_from_previous_configs(
                previous_configs,
                n_init_points,
                additional_start_points,
            )

        return sampled_points + init_points

    def _get_init_points_from_previous_configs(
        self,
        previous_configs: list[Configuration],
        n_points: int,
        additional_start_points: list[tuple[float, Configuration]] | None,
    ) -> list[Configuration]:
        """
        Generate a set of initial points from the previous configurations and possibly additional points.

        The idea is to decouple runhistory from the local search model and replace it with a more general
        form (list[Configuration]). This is useful to more quickly collect new configurations
        along the iterations, rather than feeding it to the runhistory every time.

        create three lists and concatenate them:
        1. sorted the previous configs by acquisition value
        2. sorted the previous configs by marginal predictive costs
        3. additional start points

        and create a list that carries unique configurations only. Crucially,
        when reading from left to right, all but the first occurrence of a configuration
        are dropped.

        Parameters
        ----------
        previous_configs: list[Configuration]
            Previous configuration (e.g., from the runhistory).
        n_points: int
            Number of initial points to be generated; selected from previous configs (+ random configs if necessary).
        additional_start_points: list[tuple[float, Configuration]] | None
            Additional starting points.

        Returns
        -------
        init_points: list[Configuration]
            A list of initial points.
        """
        assert self._acquisition_function is not None

        # configurations with the lowest predictive cost, check for None to make unit tests work
        if self._acquisition_function.model is not None:
            conf_array = convert_configurations_to_array(previous_configs)
            costs = self._acquisition_function.model.predict_marginalized(conf_array)[0]
            assert len(conf_array) == len(costs), (conf_array.shape, costs.shape)

            # In case of the predictive model returning the prediction for more than one objective per configuration
            # (for example multi-objective or EIPS) it is not immediately clear how to sort according to the cost
            # of a configuration. Therefore, we simply follow the ParEGO approach and use a random scalarization.
            if len(costs.shape) == 2 and costs.shape[1] > 1:
                weights = np.array([self._rng.rand() for _ in range(costs.shape[1])])
                weights = weights / np.sum(weights)
                costs = costs @ weights

            # From here: make argsort result to be random between equal values
            # http://stackoverflow.com/questions/20197990/how-to-make-argsort-result-to-be-random-between-equal-values
            random = self._rng.rand(len(costs))
            indices = np.lexsort((random.flatten(), costs.flatten()))  # Last column is primary sort key!

            # Cannot use zip here because the indices array cannot index the
            # rand_configs list, because the second is a pure python list
            previous_configs_sorted_by_cost = [previous_configs[ind] for ind in indices][:n_points]
        else:
            previous_configs_sorted_by_cost = []

        if additional_start_points is not None:
            additional_start_points = [asp[1] for asp in additional_start_points]
        else:
            additional_start_points = []

        init_points = []
        init_points_as_set: set[Configuration] = set()
        for cand in itertools.chain(
            previous_configs_sorted_by_cost,
            additional_start_points,
        ):
            if cand not in init_points_as_set:
                init_points.append(cand)
                init_points_as_set.add(cand)

        return init_points

    def _search(
        self,
        start_points: list[Configuration],
    ) -> list[tuple[float, Configuration]]:
        """Optimize the acquisition function.

        Execution:
        1. Neighbour generation strategy for each of the starting points is according to
        ConfigSpace.utils.get_one_exchange_neighbourhood.
        2. Each of the starting points create a local search, that can be active.
        if it is active, request a neighbour of its neightbourhood and evaluate it.
        3. Comparing the acquisition function of the neighbors with the acquisition value of the
        candidate.
        If it improved, then the candidate is replaced by the neighbor. And this candidate is
        investigated again with two new neighbours.
        If it did not improve, it is investigated with twice as many new neighbours
        (at most self._vectorization_max_obtain neighbours).
        The local search for a starting point is stopped if the number of evaluations is larger
        than self._n_steps_plateau_walk.


        Parameters
        ----------
        start_points : list[Configuration]
            Starting points for the search.

        Returns
        -------
        list[tuple[float, Configuration]]
            Candidates with their acquisition function value. (acq value, candidate)
        """
        assert self._acquisition_function is not None

        # Gather data structure for starting points
        if isinstance(start_points, Configuration):
            start_points = [start_points]

        candidates = start_points
        # Compute the acquisition value of the candidates
        num_candidates = len(candidates)
        acq_val_candidates_ = self._acquisition_function(candidates)

        if num_candidates == 1:
            acq_val_candidates = [acq_val_candidates_[0][0]]
        else:
            acq_val_candidates = [a[0] for a in acq_val_candidates_]

        # Set up additional variables required to do vectorized local search:
        # whether the i-th local search is still running
        active = [True] * num_candidates
        # number of plateau walks of the i-th local search. Reaching the maximum number is the stopping criterion of
        # the local search.
        n_no_plateau_walk = [0] * num_candidates
        # tracking the number of steps for logging purposes
        local_search_steps = [0] * num_candidates
        # tracking the number of neighbors looked at for logging purposes
        neighbors_looked_at = [0] * num_candidates
        # tracking the number of neighbors generated for logging purposse
        neighbors_generated = [0] * num_candidates
        # how many neighbors were obtained for the i-th local search. Important to map the individual acquisition
        # function values to the correct local search run
        obtain_n = [self._vectorization_min_obtain] * num_candidates
        # Tracking the time it takes to compute the acquisition function
        times = []

        # Set up the neighborhood generators
        neighborhood_iterators = []
        for i, inc in enumerate(candidates):
            neighborhood_iterators.append(
                # get_one_exchange_neighbourhood implementational details:
                # https://github.com/automl/ConfigSpace/blob/05ab3da2a06c084ba920e8e4e3f62f2e87e81442/ConfigSpace/util.pyx#L95
                # Return all configurations in a one-exchange neighborhood.
                #
                #     The method is implemented as defined by:
                #     Frank Hutter, Holger H. Hoos and Kevin Leyton-Brown
                #     Sequential Model-Based Optimization for General Algorithm Configuration
                #     In Proceedings of the conference on Learning and Intelligent
                #     Optimization(LION 5)
                get_one_exchange_neighbourhood(inc, seed=self._rng.randint(low=0, high=100000))
            )
            local_search_steps[i] += 1

        # Keeping track of configurations with equal acquisition value for plateau walking
        neighbors_w_equal_acq: list[list[Configuration]] = [[] for _ in range(num_candidates)]

        num_iters = 0
        while np.any(active):

            # If the maximum number of steps is reached, stop the local search
            if num_iters is not None and num_iters == self._max_steps:
                break

            num_iters += 1
            # Whether the i-th local search improved. When a new neighborhood is generated, this is used to determine
            # whether a step was made (improvement) or not (iterator exhausted)
            improved = [False] * num_candidates
            # Used to request a new neighborhood for the candidates of the i-th local search
            new_neighborhood = [False] * num_candidates

            # gather all neighbors
            neighbors = []
            for i, neighborhood_iterator in enumerate(neighborhood_iterators):
                if active[i]:
                    neighbors_for_i = []
                    for j in range(obtain_n[i]):
                        try:
                            n = next(neighborhood_iterator)
                            neighbors_generated[i] += 1
                            neighbors_for_i.append(n)
                        except ValueError as e:
                            # `neighborhood_iterator` raises `ValueError` with some probability when it reaches
                            # an invalid configuration.
                            logger.debug(e)
                            new_neighborhood[i] = True
                        except StopIteration:
                            new_neighborhood[i] = True
                            break
                    obtain_n[i] = len(neighbors_for_i)
                    neighbors.extend(neighbors_for_i)

            if len(neighbors) != 0:
                start_time = time.time()
                acq_val = self._acquisition_function(neighbors)
                end_time = time.time()
                times.append(end_time - start_time)
                if np.ndim(acq_val.shape) == 0:
                    acq_val = np.asarray([acq_val])

                # Comparing the acquisition function of the neighbors with the acquisition value of the candidate
                acq_index = 0
                # Iterating the all i local searches
                for i in range(num_candidates):
                    if not active[i]:
                        continue

                    # And for each local search we know how many neighbors we obtained
                    for j in range(obtain_n[i]):
                        # The next line is only true if there was an improvement and we basically need to iterate to
                        # the i+1-th local search
                        if improved[i]:
                            acq_index += 1
                        else:
                            neighbors_looked_at[i] += 1

                            # Found a better configuration
                            if acq_val[acq_index] > acq_val_candidates[i]:
                                is_valid = False
                                try:
                                    neighbors[acq_index].check_valid_configuration()
                                    is_valid = True
                                except (ValueError, ForbiddenValueError) as e:
                                    logger.debug("Local search %d: %s", i, e)

                                if is_valid:
                                    # We comment this as it just spams the log
                                    # logger.debug(
                                    #     "Local search %d: Switch to one of the neighbors (after %d configurations).",
                                    #     i,
                                    #     neighbors_looked_at[i],
                                    # )
                                    candidates[i] = neighbors[acq_index]
                                    acq_val_candidates[i] = acq_val[acq_index]
                                    new_neighborhood[i] = True
                                    improved[i] = True
                                    local_search_steps[i] += 1
                                    neighbors_w_equal_acq[i] = []
                                    obtain_n[i] = 1
                            # Found an equally well performing configuration, keeping it for plateau walking
                            elif acq_val[acq_index] == acq_val_candidates[i]:
                                neighbors_w_equal_acq[i].append(neighbors[acq_index])

                            acq_index += 1

            # Now we check whether we need to create new neighborhoods and whether we need to increase the number of
            # plateau walks for one of the local searches. Also disables local searches if the number of plateau walks
            # is reached (and all being switched off is the termination criterion).
            for i in range(num_candidates):
                if not active[i]:
                    continue

                if obtain_n[i] == 0 or improved[i]:
                    obtain_n[i] = 2
                else:
                    obtain_n[i] = obtain_n[i] * 2
                    obtain_n[i] = min(obtain_n[i], self._vectorization_max_obtain)

                if new_neighborhood[i]:
                    if not improved[i] and n_no_plateau_walk[i] < self._n_steps_plateau_walk:
                        if len(neighbors_w_equal_acq[i]) != 0:
                            candidates[i] = neighbors_w_equal_acq[i][0]
                            neighbors_w_equal_acq[i] = []
                        n_no_plateau_walk[i] += 1
                    if n_no_plateau_walk[i] >= self._n_steps_plateau_walk:
                        active[i] = False
                        continue

                    neighborhood_iterators[i] = get_one_exchange_neighbourhood(
                        candidates[i],
                        seed=self._rng.randint(low=0, high=100000),
                    )

        logger.debug(
            "Local searches took %s steps and looked at %s configurations. Computing the acquisition function in "
            "vectorized for took %f seconds on average.",
            local_search_steps,
            neighbors_looked_at,
            np.mean(times),
        )

        return [(a, i) for a, i in zip(acq_val_candidates, candidates)]



from distutils.spawn import find_executable

import matplotlib

matplotlib.use("Agg")
import base64
import io

import matplotlib.pyplot as plt

# IEEETrans double column standard
FIG_WIDTH = 252.0 / 72.27  # 1pt is 1/72.27 inches
FIG_HEIGHT = FIG_WIDTH / 1.618  # golden ratio


class StyledPlot:
    """
    Overwrites default settings from matplotlib.pyplot.
    If a function is not overwritten, the default function will be used.
    """

    def __init__(self):
        plt.style.use("seaborn")

        # Set MatPlotLib defaults
        if find_executable("latex"):
            plt.rcParams.update(
                {
                    "text.usetex": True,
                    "font.family": "serif",
                    "font.serif": ["Computer Modern"],
                }
            )

        plt.rc("figure", autolayout=True)
        plt.rc("xtick", labelsize=8)
        plt.rc("ytick", labelsize=8)
        plt.rc("axes", labelsize=12)
        plt.rc("axes", titlesize=12)
        plt.rc("legend", fontsize=8)

        self.plt = plt

    def figure(self, cols=1, rows=1, dpi=200):
        # Clean all
        self.plt.cla()
        self.plt.clf()

        f = self.plt.figure(figsize=(FIG_WIDTH * cols, FIG_HEIGHT * rows), dpi=dpi)
        f.tight_layout()

        return f

    def save_figure(self, filename):
        self.plt.savefig(filename, dpi=400, bbox_inches="tight")
        self.plt.close()

    def render(self):
        # Ccreate a virtual file which matplotlib can use to save the figure
        buffer = io.BytesIO()
        self.plt.savefig(buffer, dpi=400, bbox_inches="tight")
        buffer.seek(0)

        # Display any kind of image taken from
        # https://github.com/plotly/dash/issues/71
        encoded_image = base64.b64encode(buffer.read())
        return "data:image/png;base64,{}".format(encoded_image.decode())

    def xlim(self, xmin, xmax):
        xmin_with_margin = xmin - 0.05 * (xmax - xmin)
        xmax_with_margin = xmax + 0.05 * (xmax - xmin)
        self.plt.xlim(xmin_with_margin, xmax_with_margin)

    def ylim(self, ymin, ymax, margin=True):
        if margin:
            ymin_with_margin = ymin - 0.05 * (ymax - ymin)
            ymax_with_margin = ymax + 0.05 * (ymax - ymin)
            self.plt.ylim(ymin_with_margin, ymax_with_margin)
        else:
            self.plt.ylim(ymin, ymax)

    # def grid(self):
    #    pass
    #    #self.plt.grid(b=True, color='black', linestyle='--', linewidth=0.5, axis='y', zorder=0, alpha=0.5)

    def boxplot(self, values, positions, color, widths=0.5):
        bp = self.plt.boxplot(values, positions=positions, patch_artist=True, widths=widths)

        for box in bp["boxes"]:
            box.set_facecolor(color)
            box.set(linewidth=0.3)

        for whisker in bp["whiskers"]:
            whisker.set(color=color, linewidth=0.5)

        for cap in bp["caps"]:
            cap.set(color=color, linewidth=0.5)

        for median in bp["medians"]:
            median.set(color="black", linewidth=0.5)

        for flier in bp["fliers"]:
            flier.set(
                marker="o",
                markersize=3,
                markerfacecolor=color,
                linestyle="none",
                markeredgecolor="none",
                color=color,
                alpha=0.5,
            )

    def legend(self, cols=1, loc=None, title=None, outside=False):
        kwargs = {
            "ncol": cols,
            "columnspacing": 0.8,
            "labelspacing": 0,
            "loc": loc,
            "fancybox": False,
            "framealpha": 0.8,
            "frameon": True,
            "borderaxespad": 0.4,
            "facecolor": "white",
            "title": title,
        }

        if loc is not None:
            kwargs["loc"] = loc

        if outside:
            kwargs.update({"loc": "upper left", "bbox_to_anchor": (1, 1)})

        legend = self.plt.legend(**kwargs)
        legend.set_zorder(500)

        if outside:
            legend.get_frame().set_linewidth(0.0)
            legend.get_frame().set_edgecolor("white")
        else:
            legend.get_frame().set_linewidth(0.5)
            legend.get_frame().set_edgecolor("gray")

    def get_color(self, id):
        import seaborn as sns

        pal = sns.color_palette()
        hex_codes = pal.as_hex()

        return hex_codes[id % len(hex_codes)]

    def __getattr__(self, name):
        """
        Make sure we access self.plt directly.
        """

        try:
            return self.__getattribute__(name)
        except:
            return self.plt.__getattribute__(name)


plt = StyledPlot()



from __future__ import annotations

import hashlib
import logging
from functools import partial

import numpy as np
from ConfigSpace import Configuration, ConfigurationSpace
from ConfigSpace.hyperparameters import (
    BetaFloatHyperparameter,
    BetaIntegerHyperparameter,
    CategoricalHyperparameter,
    Constant,
    IntegerHyperparameter,
    NormalFloatHyperparameter,
    NormalIntegerHyperparameter,
    NumericalHyperparameter,
    OrdinalHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
)
from ConfigSpace.util import (
    ForbiddenValueError,
    deactivate_inactive_hyperparameters,
    get_one_exchange_neighbourhood,
)

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


get_one_exchange_neighbourhood = partial(get_one_exchange_neighbourhood, stdev=0.05, num_neighbors=8)


def convert_configurations_to_array(configs: list[Configuration]) -> np.ndarray:
    """Impute inactive hyperparameters in configurations with their default.

    Parameters
    ----------
    configs : List[Configuration]
        List of configuration objects.

    Returns
    -------
    np.ndarray
    """
    return np.array([config.get_array() for config in configs], dtype=np.float64)


def get_types(
    configspace: ConfigurationSpace,
    instance_features: dict[str, list[float]] | None = None,
) -> tuple[list[int], list[tuple[float, float]]]:
    """Return the types of the hyperparameters and the bounds of the
    hyperparameters and instance features.

    Warning
    -------
    The bounds for the instance features are *not* added in this function.
    """
    # Extract types vector for rf from config space and the bounds
    types = [0] * len(list(configspace.values()))
    bounds = [(np.nan, np.nan)] * len(types)

    for i, param in enumerate(list(configspace.values())):
        parents = configspace.parents_of[param.name]
        if len(parents) == 0:
            can_be_inactive = False
        else:
            can_be_inactive = True

        if isinstance(param, (CategoricalHyperparameter)):
            n_cats = len(param.choices)
            if can_be_inactive:
                n_cats = len(param.choices) + 1
            types[i] = n_cats
            bounds[i] = (int(n_cats), np.nan)
        elif isinstance(param, (OrdinalHyperparameter)):
            n_cats = len(param.sequence)
            types[i] = 0
            if can_be_inactive:
                bounds[i] = (0, int(n_cats))
            else:
                bounds[i] = (0, int(n_cats) - 1)
        elif isinstance(param, Constant):
            # For constants we simply set types to 0 which makes it a numerical parameter
            if can_be_inactive:
                bounds[i] = (2, np.nan)
                types[i] = 2
            else:
                bounds[i] = (0, np.nan)
                types[i] = 0
            # and we leave the bounds to be 0 for now
        elif isinstance(param, UniformFloatHyperparameter):
            # Are sampled on the unit hypercube thus the bounds
            # are always 0.0, 1.0
            if can_be_inactive:
                bounds[i] = (-1.0, 1.0)
            else:
                bounds[i] = (0, 1.0)
        elif isinstance(param, UniformIntegerHyperparameter):
            if can_be_inactive:
                bounds[i] = (-1.0, 1.0)
            else:
                bounds[i] = (0, 1.0)
        elif isinstance(param, NormalFloatHyperparameter):
            if can_be_inactive:
                raise ValueError("Inactive parameters not supported for Beta and Normal Hyperparameters")

            bounds[i] = (param.lower_vectorized, param.upper_vectorized)
        elif isinstance(param, NormalIntegerHyperparameter):
            if can_be_inactive:
                raise ValueError("Inactive parameters not supported for Beta and Normal Hyperparameters")

            bounds[i] = (param.lower_vectorized, param.upper_vectorized)
        elif isinstance(param, BetaFloatHyperparameter):
            if can_be_inactive:
                raise ValueError("Inactive parameters not supported for Beta and Normal Hyperparameters")

            bounds[i] = (param.lower_vectorized, param.upper_vectorized)
        elif isinstance(param, BetaIntegerHyperparameter):
            if can_be_inactive:
                raise ValueError("Inactive parameters not supported for Beta and Normal Hyperparameters")

            bounds[i] = (param.lower_vectorized, param.upper_vectorized)
        elif not isinstance(
            param,
            (
                UniformFloatHyperparameter,
                UniformIntegerHyperparameter,
                OrdinalHyperparameter,
                CategoricalHyperparameter,
                NormalFloatHyperparameter,
                NormalIntegerHyperparameter,
                BetaFloatHyperparameter,
                BetaIntegerHyperparameter,
            ),
        ):
            raise TypeError("Unknown hyperparameter type %s" % type(param))

    if instance_features is not None:
        n_features = len(list(instance_features.values())[0])
        types = types + [0] * n_features

    return types, bounds


def get_conditional_hyperparameters(X: np.ndarray, Y: np.ndarray | None = None) -> np.ndarray:
    """Returns conditional hyperparameters if values with -1 or smaller are observed. X is used
    if Y is not specified.
    """
    # Taking care of conditional hyperparameters according to Levesque et al.
    X_cond = X <= -1

    if Y is not None:
        Y_cond = Y <= -1
    else:
        Y_cond = X <= -1

    active = ~((np.expand_dims(X_cond, axis=1) != Y_cond).any(axis=2))
    return active


def get_config_hash(config: Configuration, chars: int = 6) -> str:
    """Returns a hash of the configuration."""
    return hashlib.sha1(str(config).encode("utf-8")).hexdigest()[:chars]


def print_config_changes(
    incumbent: Configuration | None,
    challenger: Configuration | None,
    logger: logging.Logger,
) -> None:
    """Compares two configurations and prints the differences."""
    if incumbent is None or challenger is None:
        return

    inc_keys = set(incumbent.keys())
    all_keys = inc_keys.union(challenger.keys())

    lines = []
    for k in sorted(all_keys):
        inc_k = incumbent.get(k, "-inactive-")
        cha_k = challenger.get(k, "-inactive-")
        lines.append(f"--- {k}: {inc_k} -> {cha_k}" + " (unchanged)" if inc_k == cha_k else "")

    msg = "\n".join(lines)
    logger.debug(msg)


def transform_continuous_designs(
    design: np.ndarray, origin: str, configspace: ConfigurationSpace
) -> list[Configuration]:
    """Transforms the continuous designs into a discrete list of configurations.

    Parameters
    ----------
    design : np.ndarray
        Array of hyperparameters originating from the initial design strategy.
    origin : str | None, defaults to None
        Label for a configuration where it originated from.
    configspace : ConfigurationSpace

    Returns
    -------
    configs : list[Configuration]
        Continuous transformed configs.
    """
    params = configspace.get_hyperparameters()
    for idx, param in enumerate(params):
        if isinstance(param, IntegerHyperparameter):
            design[:, idx] = param._inverse_transform(param._transform(design[:, idx]))
        elif isinstance(param, NumericalHyperparameter):
            continue
        elif isinstance(param, Constant):
            design_ = np.zeros(np.array(design.shape) + np.array((0, 1)))
            design_[:, :idx] = design[:, :idx]
            design_[:, idx + 1 :] = design[:, idx:]
            design = design_
        elif isinstance(param, CategoricalHyperparameter):
            v_design = design[:, idx]
            v_design[v_design == 1] = 1 - 10**-10
            design[:, idx] = np.array(v_design * len(param.choices), dtype=int)
        elif isinstance(param, OrdinalHyperparameter):
            v_design = design[:, idx]
            v_design[v_design == 1] = 1 - 10**-10
            design[:, idx] = np.array(v_design * len(param.sequence), dtype=int)
        else:
            raise ValueError("Hyperparameter not supported when transforming a continuous design.")

    configs = []
    for vector in design:
        try:
            conf = deactivate_inactive_hyperparameters(
                configuration=None, configuration_space=configspace, vector=vector
            )
        except ForbiddenValueError:
            continue

        conf.origin = origin
        configs.append(conf)

    return configs


# def check_subspace_points(
#     X: np.ndarray,
#     cont_dims: np.ndarray | list = [],
#     cat_dims: np.ndarray | list = [],
#     bounds_cont: np.ndarray | None = None,
#     bounds_cat: list[tuple] | None = None,
#     expand_bound: bool = False,
# ) -> np.ndarray:
#     """Check which points are place inside a given subspace.

#     Parameters
#     ----------
#     X: Optional[np.ndarray(N,D)],
#         points to be checked, where D = D_cont + D_cat
#     cont_dims: Union[np.ndarray(D_cont), List]
#         which dimensions represent continuous hyperparameters
#     cat_dims: Union[np.ndarray(D_cat), List]
#         which dimensions represent categorical hyperparameters
#     bounds_cont: optional[List[Tuple]]
#         subspaces bounds of categorical hyperparameters, its length is the number of continuous hyperparameters
#     bounds_cat: Optional[List[Tuple]]
#         subspaces bounds of continuous hyperparameters, its length is the number of categorical hyperparameters
#     expand_bound: bool
#         if the bound needs to be expanded to contain more points rather than the points inside the subregion
#     Return
#     ----------
#     indices_in_ss:np.ndarray(N)
#         indices of data that included in subspaces
#     """
#     if len(X.shape) == 1:
#         X = X[np.newaxis, :]
#     if len(cont_dims) == 0 and len(cat_dims) == 0:
#         return np.ones(X.shape[0], dtype=bool)

#     if len(cont_dims) > 0:
#         if bounds_cont is None:
#             raise ValueError("bounds_cont must be given if cont_dims provided")

#         if len(bounds_cont.shape) != 2 or bounds_cont.shape[1] != 2 or bounds_cont.shape[0] != len(cont_dims):
#             raise ValueError(
#                 f"bounds_cont (with shape  {bounds_cont.shape}) should be an array with shape of"
#                 f"({len(cont_dims)}, 2)"
#             )

#         data_in_ss = np.all(X[:, cont_dims] <= bounds_cont[:, 1], axis=1) & np.all(
#             X[:, cont_dims] >= bounds_cont[:, 0], axis=1
#         )

#         if expand_bound:
#             bound_left = bounds_cont[:, 0] - np.min(X[data_in_ss][:, cont_dims] - bounds_cont[:, 0], axis=0)
#             bound_right = bounds_cont[:, 1] + np.min(bounds_cont[:, 1] - X[data_in_ss][:, cont_dims], axis=0)
#             data_in_ss = np.all(X[:, cont_dims] <= bound_right, axis=1) & np.all(X[:, cont_dims] >= bound_left,
# axis=1)
#     else:
#         data_in_ss = np.ones(X.shape[0], dtype=bool)

#     if len(cat_dims) == 0:
#         return data_in_ss
#     if bounds_cat is None:
#         raise ValueError("bounds_cat must be given if cat_dims provided")

#     if len(bounds_cat) != len(cat_dims):
#         raise ValueError(
#             f"bounds_cat ({len(bounds_cat)}) and cat_dims ({len(cat_dims)}) must have " f"the same number of elements"
#         )

#     for bound_cat, cat_dim in zip(bounds_cat, cat_dims):
#         data_in_ss &= np.in1d(X[:, cat_dim], bound_cat)

#     return data_in_ss



import automl_sphinx_theme

from smac import copyright, author, version, name
from sphinx_gallery.sorting import FileNameSortKey

# from smac.cli.cmd_reader import CMDReader


options = {
    "copyright": copyright,
    "author": author,
    "version": version,
    "versions": {
        f"v{version}": "#",
        "v2.3.0": "https://automl.github.io/SMAC3/v2.3.0/",
        "v2.2.0": "https://automl.github.io/SMAC3/v2.2.0/",
        "v2.1.0": "https://automl.github.io/SMAC3/v2.1.0/",
        "v2.0.1": "https://automl.github.io/SMAC3/v2.0.1/",
        "v2.0.0": "https://automl.github.io/SMAC3/v2.0.0/",
        "v2.0.0b1": "https://automl.github.io/SMAC3/v2.0.0b1/",
        "v2.0.0a2": "https://automl.github.io/SMAC3/v2.0.0a2/",
        "v2.0.0a1": "https://automl.github.io/SMAC3/v2.0.0a1/",
        "v1.4.0": "https://automl.github.io/SMAC3/v1.4.0/",
    },
    "name": name,
    "html_theme_options": {
        "github_url": "https://github.com/automl/SMAC3",
        "twitter_url": "https://twitter.com/automl_org?lang=de",
    },
    # "ignore_pattern": ".*pcs$|.*scenario.txt$|.*spear_qcp$",
    "sphinx_gallery_conf": {
        "plot_gallery": True,
        "within_subsection_order": FileNameSortKey,
        "filename_pattern": "/",  # We want to execute all files in `examples`
        "binder": {
            # Required keys
            "org": "automl",
            "repo": "SMAC3",
            "branch": "main",
            "binderhub_url": "https://mybinder.org",
            "dependencies": ["../.binder/apt.txt", "../.binder/requirements.txt"],
            # "filepath_prefix": "<prefix>" # A prefix to prepend to any filepaths in Binder links.
            # Jupyter notebooks for Binder will be copied to this directory (relative to built documentation root).
            "notebooks_dir": "notebooks/",
            "use_jupyter_lab": True,
            # Whether Binder links should start Jupyter Lab instead of the Jupyter Notebook interface.
        },
        "ignore_pattern": ".*7_parallelization_cluster.py$",
    },
}

automl_sphinx_theme.set_options(globals(), options)

# Write outputs
# cmd_reader = CMDReader()
# cmd_reader.write_main_options_to_doc()
# cmd_reader.write_smac_options_to_doc()
# cmd_reader.write_scenario_options_to_doc()



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Callable

from pathlib import Path

import joblib
from ConfigSpace import Configuration
from dask.distributed import Client
from typing_extensions import Literal

import smac
from smac.acquisition.function.abstract_acquisition_function import (
    AbstractAcquisitionFunction,
)
from smac.acquisition.maximizer.abstract_acquisition_maximizer import (
    AbstractAcquisitionMaximizer,
)
from smac.callback.callback import Callback
from smac.initial_design.abstract_initial_design import AbstractInitialDesign
from smac.intensifier.abstract_intensifier import AbstractIntensifier
from smac.main.config_selector import ConfigSelector
from smac.main.smbo import SMBO
from smac.model.abstract_model import AbstractModel
from smac.multi_objective.abstract_multi_objective_algorithm import (
    AbstractMultiObjectiveAlgorithm,
)
from smac.random_design.abstract_random_design import AbstractRandomDesign
from smac.runhistory.dataclasses import TrialInfo, TrialValue
from smac.runhistory.encoder.abstract_encoder import AbstractRunHistoryEncoder
from smac.runhistory.runhistory import RunHistory
from smac.runner.abstract_runner import AbstractRunner
from smac.runner.dask_runner import DaskParallelRunner
from smac.runner.target_function_runner import TargetFunctionRunner
from smac.runner.target_function_script_runner import TargetFunctionScriptRunner
from smac.scenario import Scenario
from smac.utils.logging import get_logger, setup_logging

logger = get_logger(__name__)

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class AbstractFacade:
    """Facade is an abstraction on top of the SMBO backend to organize the components of a Bayesian Optimization loop
    in a configurable and separable manner to suit the various needs of different (hyperparameter) optimization
    pipelines.

    With the exception to scenario and ``target_function``, which are expected of the user, the parameters ``model``,
    ``acquisition_function``, ``acquisition_maximizer``, ``initial_design``, ``random_design``, ``intensifier``,
    ``multi_objective_algorithm``, ``runhistory_encoder`` can either be explicitly specified in the subclasses'
    ``get_*`` methods (defining a specific BO pipeline) or be instantiated by the user to overwrite a pipeline
    components explicitly.

    Parameters
    ----------
    scenario : Scenario
        The scenario object, holding all environmental information.
    target_function : Callable | str | AbstractRunner
        This function is called internally to judge a trial's performance. If a string is passed,
        it is assumed to be a script. In this case, ``TargetFunctionScriptRunner`` is used to run the script.
    model : AbstractModel | None, defaults to None
        The surrogate model.
    acquisition_function : AbstractAcquisitionFunction | None, defaults to None
        The acquisition function.
    acquisition_maximizer : AbstractAcquisitionMaximizer | None, defaults to None
        The acquisition maximizer, deciding which configuration is most promising based on the surrogate model and
        acquisition function.
    initial_design : InitialDesign | None, defaults to None
        The sampled configurations from the initial design are evaluated before the Bayesian optimization loop starts.
    random_design : RandomDesign | None, defaults to None
        The random design is used in the acquisition maximizer, deciding whether the next configuration should be drawn
        from the acquisition function or randomly.
    intensifier : AbstractIntensifier | None, defaults to None
        The intensifier decides which trial (combination of configuration, seed, budget and instance) should be run
        next.
    multi_objective_algorithm : AbstractMultiObjectiveAlgorithm | None, defaults to None
        In case of multiple objectives, the objectives need to be interpreted so that an optimization is possible.
        The multi-objective algorithm takes care of that.
    runhistory_encoder : RunHistoryEncoder | None, defaults to None
        Based on the runhistory, the surrogate model is trained. However, the data first needs to be encoded, which
        is done by the runhistory encoder. For example, inactive hyperparameters need to be encoded or cost values
        can be log transformed.
    logging_level: int | Path | Literal[False] | None
        The level of logging (the lowest level 0 indicates the debug level). If a path is passed, a yaml file is
        expected with the logging configuration. If nothing is passed, the default logging.yml from SMAC is used.
        If False is passed, SMAC will not do any customization of the logging setup and the responsibility is left
        to the user.
    callbacks: list[Callback], defaults to []
        Callbacks, which are incorporated into the optimization loop.
    overwrite: bool, defaults to False
        When True, overwrites the run results if a previous run is found that is
        consistent in the meta data with the current setup. When False and a previous run is found that is
        consistent in the meta data, the run is continued. When False and a previous run is found that is
        not consistent in the meta data, the the user is asked for the exact behaviour (overwrite completely
        or rename old run first).
    dask_client: Client | None, defaults to None
        User-created dask client, which can be used to start a dask cluster and then attach SMAC to it. This will not
        be closed automatically and will have to be closed manually if provided explicitly. If none is provided
        (default), a local one will be created for you and closed upon completion.
    """

    def __init__(
        self,
        scenario: Scenario,
        target_function: Callable | str | AbstractRunner,
        *,
        model: AbstractModel | None = None,
        acquisition_function: AbstractAcquisitionFunction | None = None,
        acquisition_maximizer: AbstractAcquisitionMaximizer | None = None,
        initial_design: AbstractInitialDesign | None = None,
        random_design: AbstractRandomDesign | None = None,
        intensifier: AbstractIntensifier | None = None,
        multi_objective_algorithm: AbstractMultiObjectiveAlgorithm | None = None,
        runhistory_encoder: AbstractRunHistoryEncoder | None = None,
        config_selector: ConfigSelector | None = None,
        logging_level: int | Path | Literal[False] | None = None,
        callbacks: list[Callback] = None,
        overwrite: bool = False,
        dask_client: Client | None = None,
    ):
        setup_logging(logging_level)

        if callbacks is None:
            callbacks = []

        if model is None:
            model = self.get_model(scenario)

        if acquisition_function is None:
            acquisition_function = self.get_acquisition_function(scenario)

        if acquisition_maximizer is None:
            acquisition_maximizer = self.get_acquisition_maximizer(scenario)

        if initial_design is None:
            initial_design = self.get_initial_design(scenario)

        if random_design is None:
            random_design = self.get_random_design(scenario)

        if intensifier is None:
            intensifier = self.get_intensifier(scenario)

        if multi_objective_algorithm is None and scenario.count_objectives() > 1:
            multi_objective_algorithm = self.get_multi_objective_algorithm(scenario=scenario)

        if runhistory_encoder is None:
            runhistory_encoder = self.get_runhistory_encoder(scenario)

        if config_selector is None:
            config_selector = self.get_config_selector(scenario)

        # Initialize empty stats and runhistory object
        runhistory = RunHistory(multi_objective_algorithm=multi_objective_algorithm)

        # Set the seed for configuration space
        scenario.configspace.seed(scenario.seed)

        # Set variables globally
        self._scenario = scenario
        self._model = model
        self._acquisition_function = acquisition_function
        self._acquisition_maximizer = acquisition_maximizer
        self._initial_design = initial_design
        self._random_design = random_design
        self._intensifier = intensifier
        self._multi_objective_algorithm = multi_objective_algorithm
        self._runhistory = runhistory
        self._runhistory_encoder = runhistory_encoder
        self._config_selector = config_selector
        self._callbacks = callbacks
        self._overwrite = overwrite

        # Prepare the algorithm executer
        runner: AbstractRunner
        if isinstance(target_function, AbstractRunner):
            runner = target_function
        elif isinstance(target_function, str):
            runner = TargetFunctionScriptRunner(
                scenario=scenario,
                target_function=target_function,
                required_arguments=self._get_signature_arguments(),
            )
        else:
            runner = TargetFunctionRunner(
                scenario=scenario,
                target_function=target_function,
                required_arguments=self._get_signature_arguments(),
            )

        # In case of multiple jobs, we need to wrap the runner again using DaskParallelRunner
        if (n_workers := scenario.n_workers) > 1 or dask_client is not None:
            if dask_client is not None and n_workers > 1:
                logger.warning(
                    "Provided `dask_client`. Ignore `scenario.n_workers`, directly set `n_workers` in `dask_client`."
                )
            else:
                available_workers = joblib.cpu_count()
                if n_workers > available_workers:
                    logger.info(f"Workers are reduced to {n_workers}.")
                    n_workers = available_workers

            # We use a dask runner for parallelization
            runner = DaskParallelRunner(single_worker=runner, dask_client=dask_client)

        # Set the runner to access it globally
        self._runner = runner

        # Adding dependencies of the components
        self._update_dependencies()

        # We have to update our meta data (basically arguments of the components)
        self._scenario._set_meta(self.meta)

        # We have to validate if the object compositions are correct and actually make sense
        self._validate()

        # Finally we configure our optimizer
        self._optimizer = self._get_optimizer()
        assert self._optimizer

        # Register callbacks here
        for callback in callbacks:
            self._optimizer.register_callback(callback)

        # Additionally, we register the runhistory callback from the intensifier to efficiently update our incumbent
        # every time new information are available
        self._optimizer.register_callback(self._intensifier.get_callback(), index=0)

    @property
    def scenario(self) -> Scenario:
        """The scenario object which holds all environment information."""
        return self._scenario

    @property
    def runhistory(self) -> RunHistory:
        """The runhistory which is filled with all trials during the optimization process."""
        return self._optimizer._runhistory

    @property
    def optimizer(self) -> SMBO:
        """The optimizer which is responsible for the BO loop. Keeps track of useful information like status."""
        return self._optimizer

    @property
    def intensifier(self) -> AbstractIntensifier:
        """The optimizer which is responsible for the BO loop. Keeps track of useful information like status."""
        return self._intensifier

    @property
    def meta(self) -> dict[str, Any]:
        """Generates a hash based on all components of the facade. This is used for the run name or to determine
        whether a run should be continued or not.
        """
        multi_objective_algorithm_meta = None
        if self._multi_objective_algorithm is not None:
            multi_objective_algorithm_meta = self._multi_objective_algorithm.meta

        meta = {
            "facade": {"name": self.__class__.__name__},
            "runner": self._runner.meta,
            "model": self._model.meta,
            "acquisition_maximizer": self._acquisition_maximizer.meta,
            "acquisition_function": self._acquisition_function.meta,
            "intensifier": self._intensifier.meta,
            "initial_design": self._initial_design.meta,
            "random_design": self._random_design.meta,
            "runhistory_encoder": self._runhistory_encoder.meta,
            "multi_objective_algorithm": multi_objective_algorithm_meta,
            "config_selector": self._config_selector.meta,
            "version": smac.version,
        }

        return meta

    def ask(self) -> TrialInfo:
        """Asks the intensifier for the next trial."""
        return self._optimizer.ask()

    def tell(self, info: TrialInfo, value: TrialValue, save: bool = True) -> None:
        """Adds the result of a trial to the runhistory and updates the intensifier.

        Parameters
        ----------
        info: TrialInfo
            Describes the trial from which to process the results.
        value: TrialValue
            Contains relevant information regarding the execution of a trial.
        save : bool, optional to True
            Whether the runhistory should be saved.
        """
        return self._optimizer.tell(info, value, save=save)

    def optimize(self, *, data_to_scatter: dict[str, Any] | None = None) -> Configuration | list[Configuration]:
        """
        Optimizes the configuration of the algorithm.

        Parameters
        ----------
        data_to_scatter: dict[str, Any] | None
            We first note that this argument is valid only dask_runner!
            When a user scatters data from their local process to the distributed network,
            this data is distributed in a round-robin fashion grouping by number of cores.
            Roughly speaking, we can keep this data in memory and then we do not have to (de-)serialize the data
            every time we would like to execute a target function with a big dataset.
            For example, when your target function has a big dataset shared across all the target function,
            this argument is very useful.

        Returns
        -------
        incumbent : Configuration
            Best found configuration.
        """
        incumbents = None
        if isinstance(data_to_scatter, dict) and len(data_to_scatter) == 0:
            raise ValueError("data_to_scatter must be None or dict with some elements, but got an empty dict.")

        try:
            incumbents = self._optimizer.optimize(data_to_scatter=data_to_scatter)
        finally:
            self._optimizer.save()

        return incumbents

    def validate(
        self,
        config: Configuration,
        *,
        seed: int | None = None,
    ) -> float | list[float]:
        """Validates a configuration on seeds different from the ones used in the optimization process and on the
        highest budget (if budget type is real-valued).

        Parameters
        ----------
        config : Configuration
            Configuration to validate
        instances : list[str] | None, defaults to None
            Which instances to validate. If None, all instances specified in the scenario are used.
            In case that the budget type is real-valued, this argument is ignored.
        seed : int | None, defaults to None
            If None, the seed from the scenario is used.

        Returns
        -------
        cost : float | list[float]
            The averaged cost of the configuration. In case of multi-fidelity, the cost of each objective is
            averaged.
        """
        return self._optimizer.validate(config, seed=seed)

    @staticmethod
    @abstractmethod
    def get_model(scenario: Scenario) -> AbstractModel:
        """Returns the surrogate cost model instance used in the BO loop."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_acquisition_function(scenario: Scenario) -> AbstractAcquisitionFunction:
        """Returns the acquisition function instance used in the BO loop,
        defining the exploration/exploitation trade-off.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_acquisition_maximizer(scenario: Scenario) -> AbstractAcquisitionMaximizer:
        """Returns the acquisition optimizer instance to be used in the BO loop,
        specifying how the acquisition function instance is optimized.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_intensifier(scenario: Scenario) -> AbstractIntensifier:
        """Returns the intensifier instance to be used in the BO loop,
        specifying how to challenge the incumbent configuration on other problem instances.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_initial_design(scenario: Scenario) -> AbstractInitialDesign:
        """Returns an instance of the initial design class to be used in the BO loop,
        specifying how the configurations the BO loop is 'warm-started' with are selected.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_random_design(scenario: Scenario) -> AbstractRandomDesign:
        """Returns an instance of the random design class to be used in the BO loop,
        specifying how to interleave the BO iterations with randomly selected configurations.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_runhistory_encoder(scenario: Scenario) -> AbstractRunHistoryEncoder:
        """Returns an instance of the runhistory encoder class to be used in the BO loop,
        specifying how the runhistory is to be prepared for the next surrogate model.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_multi_objective_algorithm(scenario: Scenario) -> AbstractMultiObjectiveAlgorithm:
        """Returns the multi-objective algorithm instance to be used in the BO loop,
        specifying the scalarization strategy for multiple objectives' costs.
        """
        raise NotImplementedError

    @staticmethod
    def get_config_selector(
        scenario: Scenario,
        *,
        retrain_after: int = 8,
        retries: int = 16,
    ) -> ConfigSelector:
        """Returns the default configuration selector."""
        return ConfigSelector(scenario, retrain_after=retrain_after, retries=retries)

    def _get_optimizer(self) -> SMBO:
        """Fills the SMBO with all the pre-initialized components."""
        return SMBO(
            scenario=self._scenario,
            runner=self._runner,
            runhistory=self._runhistory,
            intensifier=self._intensifier,
            overwrite=self._overwrite,
        )

    def _update_dependencies(self) -> None:
        """Convenience method to add some more dependencies. And ensure separable instantiation of
        the components. This is the easiest way to incorporate dependencies, although
        it might be a bit hacky.
        """
        # Set components to config selector
        self._config_selector._set_components(
            initial_design=self._initial_design,
            runhistory=self._runhistory,
            runhistory_encoder=self._runhistory_encoder,
            model=self._model,
            acquisition_function=self._acquisition_function,
            acquisition_maximizer=self._acquisition_maximizer,
            random_design=self._random_design,
            callbacks=self._callbacks,
        )

        self._runhistory_encoder.multi_objective_algorithm = self._multi_objective_algorithm
        self._runhistory_encoder.runhistory = self._runhistory
        self._acquisition_function.model = self._model
        self._acquisition_maximizer.acquisition_function = self._acquisition_function
        self._intensifier.config_selector = self._config_selector
        self._intensifier.runhistory = self._runhistory

    def _validate(self) -> None:
        """Checks if the composition is correct if there are dependencies."""
        # Make sure the same acquisition function is used
        assert self._acquisition_function == self._acquisition_maximizer._acquisition_function

    def _get_signature_arguments(self) -> list[str]:
        """Returns signature arguments, which are required by the intensifier."""
        arguments = []

        if self._intensifier.uses_seeds:
            arguments += ["seed"]

        if self._intensifier.uses_budgets:
            arguments += ["budget"]

        if self._intensifier.uses_instances:
            arguments += ["instance"]

        return arguments



from __future__ import annotations

from typing import Any

import numpy as np

from smac.random_design.abstract_random_design import AbstractRandomDesign
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class CosineAnnealingRandomDesign(AbstractRandomDesign):
    """Interleaves a random configuration according to a given probability which is decreased
    according to a cosine annealing schedule.

    Parameters
    ----------
    max_probability : float
        Initial (maximum) probability of a random configuration.
    min_probability : float
        Final (minimal) probability of a random configuration used in iteration `restart_iteration`.
    restart_iteration : int
        Restart the annealing schedule every `restart_iteration` iterations.
    seed : int
        Integer used to initialize random state.
    """

    def __init__(self, min_probability: float, max_probability: float, restart_iteration: int, seed: int = 0):
        super().__init__(seed)
        assert 0 <= min_probability <= 1
        assert 0 <= max_probability <= 1
        assert max_probability > min_probability
        assert restart_iteration > 2
        self._max_probability = max_probability
        self._min_probability = min_probability

        # Internally, iteration indices start at 0, so we need to decrease this
        self._restart_iteration = restart_iteration - 1
        self._iteration = 0
        self._probability = max_probability

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update(
            {
                "max_probability": self._max_probability,
                "min_probability": self._min_probability,
                "restart_iteration": self._restart_iteration,
            }
        )

        return meta

    def next_iteration(self) -> None:  # noqa: D102
        """Moves to the next iteration and set ``self._probability``."""
        self._iteration += 1
        if self._iteration > self._restart_iteration:
            self._iteration = 0
            logger.debug("Perform a restart.")

        self._probability = self._min_probability + (
            0.5
            * (self._max_probability - self._min_probability)
            * (1 + np.cos(self._iteration * np.pi / self._restart_iteration))
        )
        logger.debug(f"Probability for random configs: {self._probability}")

    def check(self, iteration: int) -> bool:  # noqa: D102
        assert iteration >= 0

        if self._rng.rand() <= self._probability:
            return True
        else:
            return False



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Iterator

import numpy as np
from ConfigSpace import Configuration, ConfigurationSpace

from smac.acquisition.function.abstract_acquisition_function import (
    AbstractAcquisitionFunction,
)
from smac.acquisition.maximizer.helpers import ChallengerList
from smac.random_design.abstract_random_design import AbstractRandomDesign
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class AbstractAcquisitionMaximizer:
    """Abstract class for the acquisition maximization.

    In order to use this class it has to be subclassed and the
    method `_maximize` must be implemented.

    Parameters
    ----------
    configspace : ConfigurationSpace acquisition_function : AbstractAcquisitionFunction
    challengers : int, defaults to 5000 Number of configurations sampled during the optimization process,
    details depend on the used maximizer. Also, the number of configurations that is returned by calling `maximize`.
    seed : int, defaults to 0
    """

    def __init__(
        self,
        configspace: ConfigurationSpace,
        acquisition_function: AbstractAcquisitionFunction | None = None,
        challengers: int = 5000,
        seed: int = 0,
    ):
        self._configspace = configspace
        self._acquisition_function = acquisition_function
        self._challengers = challengers
        self._seed = seed
        self._rng = np.random.RandomState(seed=seed)

    @property
    def acquisition_function(self) -> AbstractAcquisitionFunction | None:
        """The acquisition function used for maximization."""
        return self._acquisition_function

    @acquisition_function.setter
    def acquisition_function(self, acquisition_function: AbstractAcquisitionFunction) -> None:
        self._acquisition_function = acquisition_function

    @property
    def meta(self) -> dict[str, Any]:
        """Return the meta-data of the created object."""
        acquisition_function_meta = None
        if self._acquisition_function is not None:
            acquisition_function_meta = self._acquisition_function.meta

        return {
            "name": self.__class__.__name__,
            "acquisition_function": acquisition_function_meta,
            "challengers": self._challengers,
            "seed": self._seed,
        }

    def maximize(
        self,
        previous_configs: list[Configuration],
        n_points: int | None = None,
        random_design: AbstractRandomDesign | None = None,
    ) -> Iterator[Configuration]:
        """Maximize acquisition function using `_maximize`, implemented by a subclass.

        Parameters
        ----------
        previous_configs: list[Configuration]
            Previous evaluated configurations.
        n_points: int, defaults to None
            Number of points to be sampled & number of configurations to be returned. If `n_points` is not specified,
            `self._challengers` is used. Semantics depend on concrete implementation.
        random_design: AbstractRandomDesign, defaults to None
            Part of the returned ChallengerList such that we can interleave random configurations
            by a scheme defined by the random design. The method `random_design.next_iteration()`
            is called at the end of this function.

        Returns
        -------
        challengers : Iterator[Configuration]
            An iterable consisting of configurations.
        """
        if n_points is None:
            n_points = self._challengers

        def next_configs_by_acquisition_value() -> list[Configuration]:
            assert n_points is not None
            # since maximize returns a tuple of acquisition value and configuration,
            # and we only need the configuration, we return the second element of the tuple
            # for each element in the list
            return [t[1] for t in self._maximize(previous_configs, n_points)]

        challengers = ChallengerList(
            self._configspace,
            next_configs_by_acquisition_value,
            random_design,
        )

        if random_design is not None:
            random_design.next_iteration()

        return challengers

    @abstractmethod
    def _maximize(
        self,
        previous_configs: list[Configuration],
        n_points: int,
    ) -> list[tuple[float, Configuration]]:
        """Implement acquisition function maximization.

        In contrast to `maximize`, this method returns an iterable of tuples, consisting of the acquisition function
        value and the configuration. This allows to plug together different acquisition function maximizers.

        Parameters
        ----------
        previous_configs: list[Configuration]
            Previously evaluated configurations.
        n_points: int
            Number of points to be sampled.

        Returns
        -------
        challengers : list[tuple[float, Configuration]]
            A list consisting of tuples of acquisition_value and its configuration.
        """
        raise NotImplementedError()

    def _sort_by_acquisition_value(self, configs: list[Configuration]) -> list[tuple[float, Configuration]]:
        """Sort the given configurations by the acquisition value.

        Parameters
        ----------
        configs : list[Configuration]

        Returns
        -------
        challengers : list[tuple[float, Configuration]]
            Candidates ordered by their acquisition value (descending).
        """
        assert self._acquisition_function is not None
        acq_values = self._acquisition_function(configs)

        # From here
        # http://stackoverflow.com/questions/20197990/how-to-make-argsort-result-to-be-random-between-equal-values
        random = self._rng.rand(len(acq_values))

        # Last column is primary sort key!
        indices = np.lexsort((random.flatten(), acq_values.flatten()))

        # Cannot use zip here because the indices array cannot index the
        # rand_configs list, because the second is a pure python list
        return [(acq_values[ind][0], configs[ind]) for ind in indices[::-1]]



from __future__ import annotations

from typing import Any

import numpy as np
from scipy.stats import norm

from smac.acquisition.function.abstract_acquisition_function import (
    AbstractAcquisitionFunction,
)
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class EI(AbstractAcquisitionFunction):
    r"""The Expected Improvement (EI) criterion is used to decide where to evaluate a function f(x) next. The goal is to
    balance exploration and exploitation. Expected Improvement (with or without function values in log space)
    acquisition function

    :math:`EI(X) := \mathbb{E}\left[ \max\{0, f(\mathbf{X^+}) - f_{t+1}(\mathbf{X}) - \xi \} \right]`,
    with :math:`f(X^+)` as the best location.

    Reference for EI: Jones, D.R. and Schonlau, M. and Welch, W.J. (1998). Efficient Global Optimization of Expensive
    Black-Box Functions. Journal of Global Optimization 13, 455–492

    Reference for logEI: Hutter, F. and Hoos, H. and Leyton-Brown, K. and Murphy, K. (2009). An experimental
    investigation of model-based parameter optimisation: SPO and beyond. In: Conference on Genetic and
    Evolutionary Computation

    The logEI implemententation is based on the derivation of the orginal equation by:
    Watanabe, S. (2024). Derivation of Closed Form of Expected Improvement for Gaussian Process Trained on
    Log-Transformed Objective. https://arxiv.org/abs/2411.18095

    Parameters
    ----------
    xi : float, defaults to 0.0
        Controls the balance between exploration and exploitation of the
        acquisition function.
    log : bool, defaults to False
        Whether the function values are in log-space.


    Attributes
    ----------
    _xi : float
        Exploration-exloitation trade-off parameter.
    _log: bool
        Function values in log-space or not.
    _eta : float
        Current incumbent function value (best value observed so far).

    """

    def __init__(
        self,
        xi: float = 0.0,
        log: bool = False,
    ) -> None:
        super(EI, self).__init__()

        self._xi: float = xi
        self._log: bool = log
        self._eta: float | None = None

    @property
    def name(self) -> str:  # noqa: D102
        return "Expected Improvement"

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update(
            {
                "xi": self._xi,
                "log": self._log,
            }
        )

        return meta

    def _update(self, **kwargs: Any) -> None:
        """Update acsquisition function attributes

        Parameters
        ----------
        eta : float
            Function value of current incumbent.
        xi : float, optional
            Exploration-exploitation trade-off parameter
        """
        assert "eta" in kwargs
        self._eta = kwargs["eta"]

        if "xi" in kwargs and kwargs["xi"] is not None:
            self._xi = kwargs["xi"]

    def _compute(self, X: np.ndarray) -> np.ndarray:
        """Compute EI acquisition value

        Parameters
        ----------
        X : np.ndarray [N, D]
            The input points where the acquisition function should be evaluated. The dimensionality of X is (N, D),
            with N as the number of points to evaluate at and D is the number of dimensions of one X.

        Returns
        -------
        np.ndarray [N,1]
            Acquisition function values wrt X.

        Raises
        ------
        ValueError
            If `update` has not been called before (current incumbent value `eta` unspecified).
        ValueError
            If EI is < 0 for at least one sample (normal function value space).
        ValueError
            If EI is < 0 for at least one sample (log function value space).
        """
        assert self._model is not None
        assert self._xi is not None

        if self._eta is None:
            raise ValueError(
                "No current best specified. Call update("
                "eta=<int>) to inform the acquisition function "
                "about the current best value."
            )

        if not self._log:
            if len(X.shape) == 1:
                X = X[:, np.newaxis]

            m, v = self._model.predict_marginalized(X)
            s = np.sqrt(v)

            def calculate_f() -> np.ndarray:
                z = (self._eta - m - self._xi) / s
                return (self._eta - m - self._xi) * norm.cdf(z) + s * norm.pdf(z)

            if np.any(s == 0.0):
                # if std is zero, we have observed x on all instances
                # using a RF, std should be never exactly 0.0
                # Avoid zero division by setting all zeros in s to one.
                # Consider the corresponding results in f to be zero.
                logger.warning("Predicted std is 0.0 for at least one sample.")
                s_copy = np.copy(s)
                s[s_copy == 0.0] = 1.0
                f = calculate_f()
                f[s_copy == 0.0] = 0.0
            else:
                f = calculate_f()

            if (f < 0).any():
                raise ValueError("Expected Improvement is smaller than 0 for at least one " "sample.")

            return f
        else:
            if len(X.shape) == 1:
                X = X[:, np.newaxis]

            m, var_ = self._model.predict_marginalized(X)
            std = np.sqrt(var_)

            def calculate_log_ei() -> np.ndarray:
                # we expect that f_min is in log-space
                assert self._eta is not None
                assert self._xi is not None

                f_min = self._eta - self._xi
                v = (f_min - m) / std
                return (np.exp(f_min) * norm.cdf(v)) - (np.exp(0.5 * var_ + m) * norm.cdf(v - std))

            if np.any(std == 0.0):
                # if std is zero, we have observed x on all instances
                # using a RF, std should be never exactly 0.0
                # Avoid zero division by setting all zeros in s to one.
                # Consider the corresponding results in f to be zero.
                logger.warning("Predicted std is 0.0 for at least one sample.")
                std_copy = np.copy(std)
                std[std_copy == 0.0] = 1.0
                log_ei = calculate_log_ei()
                log_ei[std_copy == 0.0] = 0.0
            else:
                log_ei = calculate_log_ei()

            if (log_ei < 0).any():
                raise ValueError("Expected Improvement is smaller than 0 for at least one sample.")

            return log_ei.reshape((-1, 1))


class EIPS(EI):
    r"""Expected Improvement per Second acquisition function

    :math:`EI(X) := \frac{\mathbb{E}\left[\max\{0,f(\mathbf{X^+})-f_{t+1}(\mathbf{X})-\xi\right]\}]}{np.log(r(x))}`,
    with :math:`f(X^+)` as the best location and :math:`r(x)` as runtime.

    Parameters
    ----------
    xi : float, defaults to 0.0
        Controls the balance between exploration and exploitation of the acquisition function.
    """

    def __init__(self, xi: float = 0.0) -> None:
        super(EIPS, self).__init__(xi=xi)

    @property
    def name(self) -> str:  # noqa: D102
        return "Expected Improvement per Second"

    def _compute(self, X: np.ndarray) -> np.ndarray:
        """Compute EI per second acquisition value

        Parameters
        ----------
        X : np.ndarray [N, D]
            The input points where the acquisition function should be evaluated. The dimensionality of X is (N, D),
            with N as the number of points to evaluate at and D is the number of dimensions of one X.

        Returns
        -------
        np.ndarray [N,1]
            Acquisition function values wrt X.

        Raises
        ------
        ValueError
            If the mean has the wrong shape, should have shape (-1, 2).
        ValueError
            If the variance has the wrong shape, should have shape (-1, 2).
        ValueError
            If `update` has not been called before (current incumbent value `eta` unspecified).
        ValueError
            If EIPS is < 0 for at least one sample.
        """
        assert self._model is not None
        if len(X.shape) == 1:
            X = X[:, np.newaxis]

        m, v = self._model.predict_marginalized(X)
        if m.shape[1] != 2:
            raise ValueError(f"m has wrong shape: {m.shape} != (-1, 2)")
        if v.shape[1] != 2:
            raise ValueError(f"v has wrong shape: {v.shape} != (-1, 2)")

        m_cost = m[:, 0]
        v_cost = v[:, 0]

        # The model already predicts log(runtime)
        m_runtime = m[:, 1]
        s = np.sqrt(v_cost)

        if self._eta is None:
            raise ValueError(
                "No current best specified. Call update("
                "eta=<int>) to inform the acquisition function "
                "about the current best value."
            )

        def calculate_f() -> np.ndarray:
            z = (self._eta - m_cost - self._xi) / s
            f = (self._eta - m_cost - self._xi) * norm.cdf(z) + s * norm.pdf(z)
            f = f / m_runtime

            return f

        if np.any(s == 0.0):
            # if std is zero, we have observed x on all instances
            # using a RF, std should be never exactly 0.0
            # Avoid zero division by setting all zeros in s to one.
            # Consider the corresponding results in f to be zero.
            logger.warning("Predicted std is 0.0 for at least one sample.")
            s_copy = np.copy(s)
            s[s_copy == 0.0] = 1.0
            f = calculate_f()
            f[s_copy == 0.0] = 0.0
        else:
            f = calculate_f()

        if (f < 0).any():
            raise ValueError("Expected Improvement per Second is smaller than 0 " "for at least one sample.")

        return f.reshape((-1, 1))



from __future__ import annotations

from abc import abstractmethod
from typing import Any

import numpy as np
from ConfigSpace import Configuration

from smac.model.abstract_model import AbstractModel
from smac.utils.configspace import convert_configurations_to_array
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


logger = get_logger(__name__)


class AbstractAcquisitionFunction:
    """Abstract base class for acquisition function."""

    def __init__(self) -> None:
        self._model: AbstractModel | None = None

    @property
    def name(self) -> str:
        """Returns the full name of the acquisition function."""
        raise NotImplementedError

    @property
    def meta(self) -> dict[str, Any]:
        """Returns the meta data of the created object."""
        return {
            "name": self.__class__.__name__,
        }

    @property
    def model(self) -> AbstractModel | None:
        """Return the used surrogate model in the acquisition function."""
        return self._model

    @model.setter
    def model(self, model: AbstractModel) -> None:
        """Updates the surrogate model."""
        self._model = model

    def update(self, model: AbstractModel, **kwargs: Any) -> None:
        """Update the acquisition function attributes required for calculation.

        This method will be called after fitting the model, but before maximizing the acquisition
        function. As an examples, EI uses it to update the current fmin. The default implementation only updates the
        attributes of the acquisition function which are already present.

        Calls `_update` to update the acquisition function attributes.

        Parameters
        ----------
        model : AbstractModel
            The model which was used to fit the data.
        kwargs : Any
            Additional arguments to update the specific acquisition function.
        """
        self.model = model
        self._update(**kwargs)

    def _update(self, **kwargs: Any) -> None:
        """Update acsquisition function attributes

        Might be different for each child class.
        """
        pass

    def __call__(self, configurations: list[Configuration]) -> np.ndarray:
        """Compute the acquisition value for a given configuration.

        Parameters
        ----------
        configurations : list[Configuration]
            The configurations where the acquisition function should be evaluated.

        Returns
        -------
        np.ndarray [N, 1]
            Acquisition values for X
        """
        X = convert_configurations_to_array(configurations)
        if len(X.shape) == 1:
            X = X[np.newaxis, :]

        acq = self._compute(X)
        if np.any(np.isnan(acq)):
            idx = np.where(np.isnan(acq))[0]
            acq[idx, :] = -np.finfo(float).max

        return acq

    @abstractmethod
    def _compute(self, X: np.ndarray) -> np.ndarray:
        """Compute the acquisition value for a given point X. This function has to be overwritten
        in a derived class.

        Parameters
        ----------
        X : np.ndarray [N, D]
            The input points where the acquisition function should be evaluated. The dimensionality of X is (N, D),
            with N as the number of points to evaluate at and D is the number of dimensions of one X.

        Returns
        -------
        np.ndarray [N,1]
            Acquisition function values wrt X.
        """
        raise NotImplementedError



# from __future__ import annotations

# import logging
# import warnings
# from collections import OrderedDict
# from typing import Any, TypeVar


# import gpytorch
# import numpy as np
# from smac.model.gaussian_process.utils.gpytorch import ExactGaussianProcessModel
# import torch
# from botorch.optim.numpy_converter import module_to_array, set_params_with_array  # noqa
# from botorch.optim.utils import _get_extra_mll_args, _scipy_objective_and_grad  # noqa
# from gpytorch.constraints.constraints import Interval  # noqa
# from gpytorch.distributions.multivariate_normal import MultivariateNormal  # noqa
# from gpytorch.kernels import Kernel  # noqa
# from gpytorch.likelihoods import FixedNoiseGaussianLikelihood, GaussianLikelihood  # noqa
# from gpytorch.means import ZeroMean  # noqa
# from gpytorch.mlls import ExactMarginalLogLikelihood  # noqa
# from gpytorch.models import ExactGP  # noqa
# from gpytorch.priors import HorseshoePrior  # noqa
# from gpytorch.utils.errors import NanError, NotPSDError  # noqa
# from scipy import optimize
# from scipy.stats.qmc import LatinHypercube  # noqa
# import logging
# import warnings
# from collections import OrderedDict
# from typing import Any, TypeVar

# """
# import gpytorch
# import numpy as np
# import torch
# from botorch.optim.numpy_converter import module_to_array, set_params_with_array
# from botorch.optim.utils import _get_extra_mll_args, _scipy_objective_and_grad
# from gpytorch.constraints.constraints import Interval
# from gpytorch.distributions.multivariate_normal import MultivariateNormal
# from gpytorch.kernels import Kernel
# from gpytorch.likelihoods import FixedNoiseGaussianLikelihood, GaussianLikelihood
# from gpytorch.means import ZeroMean
# from gpytorch.mlls import ExactMarginalLogLikelihood
# from gpytorch.models import ExactGP
# from gpytorch.priors import HorseshoePrior
# from gpytorch.utils.errors import NanError, NotPSDError
# from scipy import optimize
# from scipy.stats.qmc import LatinHypercube
# """

# from ConfigSpace import ConfigurationSpace
# from smac.constants import VERY_SMALL_NUMBER
# from smac.model.gaussian_process.abstract_gaussian_process import AbstractGaussianProcess
# from smac.model.utils import check_subspace_points
# from smac.model.gaussian_process.kernels._boing import FITCKernel, FITCMean


# # from smac.model.utils import check_subspace_points  # noqa
# from smac.model.gaussian_process.kernels._boing import FITCKernel, FITCMean  # noqa

# warnings.filterwarnings("ignore", module="gpytorch")

# logger = logging.getLogger(__name__)


# Self = TypeVar("Self", bound="GPyTorchGaussianProcess")


# class GPyTorchGaussianProcess(AbstractGaussianProcess):
#     """A Gaussian process written with GPyTorch. The interface is written to be compatible with partial sparse
# gaussian
#     process.

#     Parameters
#     ----------
#     configspace : ConfigurationSpace
#     kernel : Kernel
#         Kernel which is used for the Gaussian process.
#     n_restarts : int, defaults to 10
#         Number of restarts for the Gaussian process hyperparameter optimization.
#     normalize_y : bool, defaults to True
#         Zero mean unit variance normalization of the output values.
#     likelihood : FixedNoiseGaussianLikelihood | None, defaults to None
#         The Gaussian likelihood (or noise).
#     instance_features : dict[str, list[int | float]] | None, defaults to None
#         Features (list of int or floats) of the instances (str). The features are incorporated into the X data,
#         on which the model is trained on.
#     pca_components : float, defaults to 7
#         Number of components to keep when using PCA to reduce dimensionality of instance features.
#     seed : int
#     """

#     def __init__(
#         self,
#         configspace: ConfigurationSpace,
#         kernel: Kernel,
#         n_restarts: int = 10,
#         normalize_y: bool = True,
#         likelihood: FixedNoiseGaussianLikelihood | None = None,
#         instance_features: dict[str, list[int | float]] | None = None,
#         pca_components: int | None = 7,
#         seed: int = 0,
#     ):
#         if n_restarts <= 0:
#             raise ValueError("The argument `n_restarts` needs to be positive.")

#         super(GPyTorchGaussianProcess, self).__init__(
#             configspace=configspace,
#             kernel=kernel,
#             instance_features=instance_features,
#             pca_components=pca_components,
#             seed=seed,
#         )

#         if likelihood is None:
#             noise_prior = HorseshoePrior(0.1)
#             likelihood = GaussianLikelihood(
#                 noise_prior=noise_prior,
#                 noise_constraint=Interval(
#                     np.exp(-25),
#                     np.exp(2),
#                     transform=None,
#                 ),
#             ).double()

#         self._likelihood = likelihood
#         self._normalize_y = normalize_y
#         self._n_restarts = n_restarts
#         self._hypers = np.empty((0,))
#         self._property_dict: OrderedDict = OrderedDict()
#         self._is_trained = False

#     @property
#     def meta(self) -> dict[str, Any]:  # noqa: D102
#         meta = super().meta
#         meta.update(
#             {
#                 "n_restarts": self._n_restarts,
#                 "normalize_y": self._normalize_y,
#                 "likelihood": self._likelihood.__data__,
#             }
#         )

#         return meta

#     def _train(
#         self: Self,
#         X: np.ndarray,
#         y: np.ndarray,
#         optimize_hyperparameters: bool = True,
#     ) -> Self:
#         """Computes the Cholesky decomposition of the covariance of X and estimates the GP
#         hyperparameters by optimizing the marginal log likelihood. The prior mean of the GP is set to
#         the empirical mean of X.

#         Parameters
#         ----------
#         X : np.ndarray [#samples, #hyperparameter + #features]
#             Input data points.
#         Y : np.ndarray [#samples, #objectives]
#             The corresponding target values.
#         optimize_hyperparameters: boolean
#             If set to true the hyperparameters are optimized otherwise the default hyperparameters of the kernel are
#             used.
#         """
#         X = self._impute_inactive(X)

#         if len(y.shape) == 1:
#             self._n_objectives = 1
#         else:
#             self._n_objectives = y.shape[1]

#         if self._n_objectives == 1:
#             y = y.flatten()

#         if self._normalize_y:
#             y = self._normalize(y)

#         n_tries = 10
#         for i in range(n_tries):
#             try:
#                 self._gp = self._get_gaussian_process(X, y)
#                 break
#             except Exception as e:
#                 if i == n_tries - 1:
#                     # To avoid Endless loop, we need to stop it when we have n_tries unsuccessful tries.
#                     raise e

#         if optimize_hyperparameters:
#             self._hypers = self._optimize()
#             self._gp = set_params_with_array(self._gp, self._hypers, self._property_dict)
#         else:
#             self._hypers, self._property_dict, _ = module_to_array(module=self._gp)

#         self._is_trained = True
#         return self

#     def _get_gaussian_process(
#         self, X: np.ndarray | None = None, y: np.ndarray | None = None
#     ) -> ExactMarginalLogLikelihood | None:
#         """
#         Get the GP model with the given X and y values. As GPyTorch requires the input data to initialize a new
#         model, we also pass X and y here. X and y are set optional to ensure compatibility.

#         Parameters
#         ----------
#         X : np.ndarray [#samples, #hyperparameter + #features]
#             Input data points.
#         Y : np.ndarray [#samples, #objectives]
#             The corresponding target values.

#         Returns
#         -------
#         mll : ExactMarginalLogLikelihood | None
#             A GPyTorch model with zero mean and user specified covariance.
#         """
#         if X is None:
#             # To be compatible with the base model
#             return None

#         X = torch.from_numpy(X)
#         y = torch.from_numpy(y)
#         self._gp_model = ExactGaussianProcessModel(
#             X, y, likelihood=self._likelihood, base_covar_kernel=self._kernel
#         ).double()

#         mll = ExactMarginalLogLikelihood(self._likelihood, self._gp_model)
#         mll.double()

#         return mll

#     def _optimize(self) -> np.ndarray:
#         """Optimizes the marginal log likelihood and returns the best found hyperparameter
#         configuration theta.

#         Returns
#         -------
#         theta : np.ndarray
#             Hyperparameter vector that maximizes the marginal log likelihood.
#         """
#         x0, property_dict, bounds = module_to_array(module=self._gp)

#         bounds = np.asarray(bounds).transpose().tolist()

#         self._property_dict = property_dict

#         p0 = [x0]

#         # Avoid infinite sampling
#         n_tries = 5000
#         for i in range(n_tries):
#             try:
#                 gp_model = self._gp.pyro_sample_from_prior()
#                 x_out = []
#                 for key in property_dict.keys():
#                     param = gp_model
#                     param_names = key.split(".")
#                     for name in param_names:
#                         param = getattr(param, name)
#                     x_out.append(param.detach().view(-1).cpu().double().clone().numpy())
#                 sample = np.concatenate(x_out)
#                 p0.append(sample.astype(np.float64))
#             except Exception as e:
#                 if i == n_tries - 1:
#                     logger.debug(f"Fails to sample new hyperparameters because of {e}.")
#                     raise e
#                 continue
#             if len(p0) == self._n_restarts:
#                 break

#         self._gp_model.train()
#         self._likelihood.train()

#         theta_star = x0
#         f_opt_star = np.inf
#         for i, start_point in enumerate(p0):
#             try:
#                 theta, f_opt, _ = optimize.fmin_l_bfgs_b(
#                     _scipy_objective_and_grad,
#                     start_point,
#                     args=(self._gp, property_dict),
#                     bounds=bounds,
#                 )
#             except NotPSDError as e:
#                 logger.warning(f"Fail to optimize the GP hyperparameters as an Error occurs: {e}.")
#                 f_opt = np.inf
#                 theta = start_point

#             if f_opt < f_opt_star:
#                 f_opt_star = f_opt
#                 theta_star = theta

#         return theta_star

#     def _predict(
#         self,
#         X: np.ndarray,
#         covariance_type: str | None = "diagonal",
#     ) -> tuple[np.ndarray, np.ndarray | None]:
#         if not self._is_trained:
#             raise Exception("Model has to be trained first.")

#         X_test = torch.from_numpy(self._impute_inactive(X))
#         self._likelihood.eval()
#         self._gp_model.eval()

#         with torch.no_grad(), gpytorch.settings.fast_pred_var():

#             observed_pred = self._likelihood(self._gp_model(X_test))

#             mu = observed_pred.mean.numpy()
#             if covariance_type is None:
#                 var = None

#                 if self._normalize_y:
#                     mu = self._untransform_y(mu)

#             else:
#                 if covariance_type != "full":
#                     var = observed_pred.stddev.numpy()
#                     var = var**2  # since we get standard deviation for faster computation
#                 else:
#                     # output full covariance
#                     var = observed_pred.covariance_matrix.numpy()

#                 # Clip negative variances and set them to the smallest
#                 # positive float value
#                 var = np.clip(var, VERY_SMALL_NUMBER, np.inf)

#                 if self._normalize_y:
#                     mu, var = self._untransform_y(mu, var)

#                 if covariance_type == "diagonal":
#                     var = np.sqrt(var)  # converting variance to std deviation if specified

#         return mu, var

#     def sample_functions(self, X_test: np.ndarray, n_funcs: int = 1) -> np.ndarray:
#         """Samples F function values from the current posterior at the N specified test points.

#         Parameters
#         ----------
#         X : np.ndarray [#samples, #hyperparameter + #features]
#             Input data points.
#         n_funcs: int
#             Number of function values that are drawn at each test point.

#         Returns
#         -------
#         function_samples : np.ndarray
#             The F function values drawn at the N test points.
#         """
#         if not self._is_trained:
#             raise Exception("Model has to be trained first!")
#         self._likelihood.eval()
#         self._gp_model.eval()

#         X_test = torch.from_numpy(self._impute_inactive(X_test))
#         with torch.no_grad():
#             funcs = self._likelihood(self._gp_model(X_test)).sample(torch.Size([n_funcs])).t().cpu().numpy()

#         if self._normalize_y:
#             funcs = self._untransform_y(funcs)

#         if len(funcs.shape) == 1:
#             return funcs[None, :]
#         else:
#             return funcs


# class GloballyAugmentedLocalGaussianProcess(GPyTorchGaussianProcess):
#     def __init__(
#         self,
#         configspace: ConfigurationSpace,
#         bounds_cont: np.ndarray,
#         bounds_cat: list[tuple],
#         kernel: Kernel,
#         num_inducing_points: int = 2,
#         likelihood: GaussianLikelihood | None = None,
#         normalize_y: bool = True,
#         n_restarts: int = 10,
#         instance_features: dict[str, list[int | float]] | None = None,
#         pca_components: int | None = 7,
#         seed: int = 0,
#     ):
#         """
#         The GP hyperparameters are obtained by optimizing the marginal log-likelihood and optimized with botorch
#         We train an LGPGA in two stages:
#         In the first stage, we only train the kernel hyperparameter and thus deactivate the gradient w.r.t the
# position
#         of the inducing points.
#         In the second stage, we use the kernel hyperparameter acquired in the first stage to initialize a new
#         variational Gaussian process and only optimize its inducing points' position with natural gradients.
#         Finally, we update the position of the inducing points and use it for evaluation.


#         Parameters
#         ----------
#         bounds_cont: np.ndarray(N_cont, 2),
#            bounds of the continuous hyperparameters, store as [[0,1] * N_cont]
#         bounds_cat: List[Tuple],
#            bounds of categorical hyperparameters
#         kernel : gpytorch kernel object
#            Specifies the kernel that is used for all Gaussian Process
#         num_inducing_points: int
#            Number of inducing points
#         likelihood: Optional[GaussianLikelihood]
#            Likelihood values
#         normalize_y : bool
#            Zero mean unit variance normalization of the output values when the model is a partial sparse GP model.
#         """
#         super(GloballyAugmentedLocalGaussianProcess, self).__init__(
#             configspace=configspace,
#             kernel=kernel,
#             likelihood=likelihood,
#             normalize_y=normalize_y,
#             n_restarts=n_restarts,
#             instance_features=instance_features,
#             pca_components=pca_components,
#             seed=seed,
#         )
#         self.cont_dims = np.where(np.array(self.types) == 0)[0]
#         self.cat_dims = np.where(np.array(self.types) != 0)[0]
#         self.bounds_cont = bounds_cont
#         self.bounds_cat = bounds_cat
#         self.num_inducing_points = num_inducing_points

#     def update_attribute(self, **kwargs: Any) -> None:
#         """We update the class attribute (for instance, number of inducing points)"""
#         for key in kwargs:
#             if not hasattr(self, key):
#                 raise AttributeError(f"{self.__class__.__name__} has no attribute named {key}")
#             setattr(self, key, kwargs[key])

#     def _train(
#         self, X: np.ndarray, y: np.ndarray, do_optimize: bool = True
#     ) -> AugmentedLocalGaussianProcess | GPyTorchGaussianProcess:
#         """
#         Update the hyperparameters of the partial sparse kernel. Depending on the number of inputs inside and
#         outside the subregion, we initialize a  PartialSparseGaussianProcess or a GaussianProcessGPyTorch

#         Parameters
#         ----------
#         X: np.ndarray (N, D)
#             Input data points. The dimensionality of X is (N, D),
#             with N as the number of points and D is the number of features., N = N_in + N_out
#         y: np.ndarray (N,)
#             The corresponding target values.
#         do_optimize: boolean
#                 If set to true, the hyperparameters are optimized otherwise,
#                 the default hyperparameters of the kernel are used.
#         """
#         X = self._impute_inactive(X)
#         if len(y.shape) == 1:
#             self._n_objectives = 1
#         else:
#             self._n_objectives = y.shape[1]
#         if self._n_objectives == 1:
#             y = y.flatten()

#         ss_data_indices = check_subspace_points(
#             X,
#             cont_dims=self.cont_dims,
#             cat_dims=self.cat_dims,
#             bounds_cont=self.bounds_cont,
#             bounds_cat=self.bounds_cat,
#             expand_bound=True,
#         )

#         if np.sum(ss_data_indices) > np.shape(y)[0] - self.num_inducing_points:
#             # we initialize a vanilla GaussianProcessGPyTorch
#             if self._normalize_y:
#                 y = self._normalize_y(y)
#             self.num_points = np.shape(y)[0]
#             get_gp_kwargs = {"X_in": X, "y_in": y, "X_out": None, "y_out": None}
#         else:
#             # we initialize a PartialSparseGaussianProcess object
#             X_in = X[ss_data_indices]
#             y_in = y[ss_data_indices]
#             X_out = X[~ss_data_indices]
#             y_out = y[~ss_data_indices]
#             self.num_points = np.shape(y_in)[0]
#             if self._normalize_y:
#                 y_in = self._normalize_y(y_in)
#                 y_out = (y_out - self.mean_y_) / self.std_y_
#             get_gp_kwargs = {"X_in": X_in, "y_in": y_in, "X_out": X_out, "y_out": y_out}

#         n_tries = 10

#         for i in range(n_tries):
#             try:
#                 self._gp = self._get_gaussian_process(**get_gp_kwargs)
#                 break
#             except Exception as e:
#                 if i == n_tries - 1:
#                     raise RuntimeError(f"Fails to initialize a GP model, {e}")

#         if do_optimize:
#             self._hypers = self._optimize()
#             self._gp = set_params_with_array(self._gp, self._hypers, self._property_dict)
#             if isinstance(self._gp.model, AugmentedLocalGaussianProcess):
#                 # we optimize the position of the inducing points and thus needs to deactivate the gradient of kernel
#                 # hyperparameters
#                 lhd = LatinHypercube(d=X.shape[-1], seed=self.rng.randint(0, 1000000))

#                 inducing_points = torch.from_numpy(lhd.random(n=self.num_inducing_points))

#                 kernel = self._gp.model.base_covar
#                 var_gp = VariationalGaussianProcess(kernel, X_inducing=inducing_points)

#                 X_out_ = torch.from_numpy(X_out)
#                 y_out_ = torch.from_numpy(y_out)

#                 variational_ngd_optimizer = gpytorch.optim.NGD(
#                     var_gp.variational_parameters(), num_data=y_out_.size(0), lr=0.1
#                 )

#                 var_gp.train()
#                 likelihood = GaussianLikelihood().double()
#                 likelihood.train()

#                 mll_func = gpytorch.mlls.PredictiveLogLikelihood

#                 var_mll = mll_func(likelihood, var_gp, num_data=y_out_.size(0))

#                 for t in var_gp.variational_parameters():
#                     t.requires_grad = False

#                 x0, property_dict, bounds = module_to_array(module=var_mll)
#                 for t in var_gp.variational_parameters():
#                     t.requires_grad = True
#                 bounds = np.asarray(bounds).transpose().tolist()

#                 start_points = [x0]

#                 inducing_idx = 0

#                 inducing_size = X_out.shape[-1] * self.num_inducing_points
#                 for p_name, attrs in property_dict.items():
#                     if p_name != "model.variational_strategy.inducing_points":
#                         # Construct the new tensor
#                         if len(attrs.shape) == 0:  # deal with scalar tensors
#                             inducing_idx = inducing_idx + 1
#                         else:
#                             inducing_idx = inducing_idx + np.prod(attrs.shape)
#                     else:
#                         break
#                 while len(start_points) < 3:
#                     new_start_point = np.random.rand(*x0.shape)
#                     new_inducing_points = torch.from_numpy(lhd.random(n=self.num_inducing_points)).flatten()
#                     new_start_point[inducing_idx : inducing_idx + inducing_size] = new_inducing_points
#                     start_points.append(new_start_point)

#                 def sci_opi_wrapper(
#                     x: np.ndarray,
#                     mll: gpytorch.module,
#                     property_dict: dict,
#                     train_inputs: torch.Tensor,
#                     train_targets: torch.Tensor,
#                 ) -> tuple[float, np.ndarray]:
#                     """
#                     A modification of from botorch.optim.utils._scipy_objective_and_grad, the key difference is that
#                     we do an additional natural gradient update before computing the gradient values
#                     Parameters
#                     ----------
#                     x: np.ndarray
#                         optimizer input
#                     mll: gpytorch.module
#                         a gpytorch module whose hyperparameters are defined by x
#                     property_dict: Dict
#                         a dict describing how x is mapped to initialize mll
#                     train_inputs: torch.Tensor (N_input, D)
#                         input points of the GP model
#                     train_targets: torch.Tensor (N_input, 1)
#                         target value of the GP model
#                     Returns
#                     ----------
#                     loss: np.ndarray
#                         loss value
#                     grad: np.ndarray
#                         gradient w.r.t. the inputs
#                     ----------
#                     """
#                     # A modification of from botorch.optim.utils._scipy_objective_and_grad:
#                     # https://botorch.org/api/_modules/botorch/optim/utils.html
#                     # The key difference is that we do an additional natural gradient update here
#                     variational_ngd_optimizer.zero_grad()

#                     mll = set_params_with_array(mll, x, property_dict)
#                     mll.zero_grad()
#                     try:  # catch linear algebra errors in gpytorch
#                         output = mll.model(train_inputs)
#                         args = [output, train_targets] + _get_extra_mll_args(mll)
#                         loss = -mll(*args).sum()
#                     except RuntimeError as e:
#                         if isinstance(e, NanError) or "singular" in e.args[0]:
#                             return float("nan"), np.full_like(x, "nan")
#                         else:
#                             raise e  # pragma: nocover
#                     loss.backward()
#                     variational_ngd_optimizer.step()
#                     param_dict = OrderedDict(mll.named_parameters())
#                     grad = []
#                     for p_name in property_dict:
#                         t = param_dict[p_name].grad
#                         if t is None:
#                             # this deals with parameters that do not affect the loss
#                             grad.append(np.zeros(property_dict[p_name].shape.numel()))
#                         else:
#                             grad.append(t.detach().view(-1).cpu().double().clone().numpy())
#                     mll.zero_grad()
#                     return loss.item(), np.concatenate(grad)

#                 theta_star = x0
#                 f_opt_star = np.inf
#                 for start_point in start_points:
#                     try:
#                         theta, f_opt, res_dict = optimize.fmin_l_bfgs_b(
#                             sci_opi_wrapper,
#                             start_point,
#                             args=(var_mll, property_dict, X_out_, y_out_),
#                             bounds=bounds,
#                             maxiter=50,
#                         )
#                         if f_opt < f_opt_star:
#                             f_opt_star = f_opt
#                             theta_star = theta
#                     except Exception as e:
#                         logger.warning(f"An exception {e} occurs during the optimizaiton")

#                 start_idx = 0
#                 # modification on botorch.optim.numpy_converter.set_params_with_array as we only need to extract the
#                 # positions of inducing points
#                 for p_name, attrs in property_dict.items():
#                     if p_name != "model.variational_strategy.inducing_points":
#                         # Construct the new tensor
#                         if len(attrs.shape) == 0:  # deal with scalar tensors
#                             start_idx = start_idx + 1
#                         else:
#                             start_idx = start_idx + np.prod(attrs.shape)
#                     else:
#                         end_idx = start_idx + np.prod(attrs.shape)
#                         X_inducing = torch.tensor(
#                             theta_star[start_idx:end_idx], dtype=attrs.dtype, device=attrs.device
#                         ).view(*attrs.shape)
#                         break
#                 # set inducing points for covariance module here
#                 self._gp_model.set_augment_module(X_inducing)
#         else:
#             self._hypers, self._property_dict, _ = module_to_array(module=self._gp)

#         self._is_trained = True
#         return self

#     def _get_gaussian_process(
#         self,
#         X_in: np.ndarray | None = None,
#         y_in: np.ndarray | None = None,
#         X_out: np.ndarray | None = None,
#         y_out: np.ndarray | None = None,
#     ) -> ExactMarginalLogLikelihood | None:
#         """
#         Construct a new GP model based on the inputs
#         If both in and out are None: return an empty model
#         If only in_x and in_y are given: return a vanilla GP model
#         If in_x, in_y, out_x, out_y are given: return a partial sparse GP model.

#         Parameters
#         ----------
#         X_in: np.ndarray (N_in, D) | None
#             Input data points inside the subregion. The dimensionality of X_in is (N_in, D),
#             with N_in as the number of points inside the subregion and D is the number of features. If it is not
# given,
#             this function will return None to be compatible with the implementation of its parent class
#         y_in: np.ndarray (N_in,) | None
#             The corresponding target values inside the subregion.
#         X_out: np.ndarray (N_out, D) | None
#             Input data points outside the subregion. The dimensionality of X_out is (N_out, D). If it is not given,
# this
#         function will return a vanilla Gaussian Process
#         y_out: np.ndarray (N_out) | None
#             The corresponding target values outside the subregion.

#         Returns
#         -------
#         mll: ExactMarginalLogLikelihood
#             a gp module
#         """
#         if X_in is None:
#             return None

#         X_in = torch.from_numpy(X_in)
#         y_in = torch.from_numpy(y_in)
#         if X_out is None:
#             self._gp_model = ExactGaussianProcessModel(
#                 X_in, y_in, likelihood=self._likelihood, base_covar_kernel=self.kernel
#             ).double()
#         else:
#             X_out = torch.from_numpy(X_out)
#             y_out = torch.from_numpy(y_out)

#             self._gp_model = AugmentedLocalGaussianProcess(
#                 X_in, y_in, X_out, y_out, likelihood=self._likelihood, base_covar_kernel=self.kernel  # type:ignore
#             ).double()
#         mll = ExactMarginalLogLikelihood(self._likelihood, self._gp_model)
#         mll.double()
#         return mll


# class ExactGaussianProcessModel(ExactGP):
#     """Exact GP model that serves as a backbone for `GPyTorchGaussianProcess`."

#     Parameters
#     ----------
#     train_X: torch.tenor
#         input feature
#     train_y: torch.tensor
#         input observations
#     base_covar_kernel: Kernel
#         covariance kernel used to compute covariance matrix
#     likelihood: GaussianLikelihood
#         GP likelihood
#     """

#     def __init__(
#         self,
#         train_X: torch.Tensor,
#         train_y: torch.Tensor,
#         base_covar_kernel: Kernel,
#         likelihood: GaussianLikelihood,
#     ):
#         super(ExactGaussianProcessModel, self).__init__(train_X, train_y, likelihood)

#         # In our experiments we find that ZeroMean more robust than ConstantMean when y is normalized
#         self._mean_module = ZeroMean()
#         self._covar_module = base_covar_kernel

#     def forward(self, x: torch.Tensor) -> MultivariateNormal:
#         """Computes the posterior mean and variance."""
#         mean_x = self._mean_module(x)
#         covar_x = self._covar_module(x)

#         return MultivariateNormal(mean_x, covar_x)


# class AugmentedLocalGaussianProcess(ExactGP):
#     def __init__(
#         self,
#         X_in: torch.Tensor,
#         y_in: torch.Tensor,
#         X_out: torch.Tensor,
#         y_out: torch.Tensor,
#         likelihood: GaussianLikelihood,
#         base_covar_kernel: Kernel,
#     ):
#         """
#         An Augmented Local GP, it is trained with the points inside a subregion while its prior is augemented by the
#         points outside the subregion (global configurations)

#         Parameters
#         ----------
#         X_in: torch.Tensor (N_in, D),
#             feature vector of the points inside the subregion
#         y_in: torch.Tensor (N_in, 1),
#             observation inside the subregion
#         X_out: torch.Tensor (N_out, D),
#             feature vector  of the points outside the subregion
#         y_out:torch.Tensor (N_out, 1),
#             observation inside the subregion
#         likelihood: GaussianLikelihood,
#             likelihood of the GP (noise)
#         base_covar_kernel: Kernel,
#             Covariance Kernel
#         """
#         X_in = X_in.unsqueeze(-1) if X_in.ndimension() == 1 else X_in
#         X_out = X_out.unsqueeze(-1) if X_out.ndimension() == 1 else X_out
#         assert X_in.shape[-1] == X_out.shape[-1]

#         super(AugmentedLocalGaussianProcess, self).__init__(X_in, y_in, likelihood)

#         self._mean_module = ZeroMean()
#         self.base_covar = base_covar_kernel

#         self.X_out = X_out
#         self.y_out = y_out
#         self.augmented = False

#     def set_augment_module(self, X_inducing: torch.Tensor) -> None:
#         """
#         Set an augmentation module, which will be used later for inference

#         Parameters
#         ----------
#         X_inducing: torch.Tensor(N_inducing, D)
#            inducing points, it needs to have the same number of dimensions as X_in
#         """
#         X_inducing = X_inducing.unsqueeze(-1) if X_inducing.ndimension() == 1 else X_inducing
#         # assert X_inducing.shape[-1] == self.X_out.shape[-1]
#         self.covar_module = FITCKernel(
#             self.base_covar, X_inducing=X_inducing, X_out=self.X_out, y_out=self.y_out, likelihood=self._likelihood
#         )
#         self.mean_module = FITCMean(covar_module=self.covar_module)
#         self.augmented = True

#     def forward(self, x: torch.Tensor) -> MultivariateNormal:
#         """
#         Compute the prior values. If optimize_kernel_hps is set True in the training phases, this model degenerates to
#         a vanilla GP model with ZeroMean and base_covar as covariance matrix. Otherwise, we apply partial sparse GP
#         mean and kernels here.
#         """
#         if not self.augmented:
#             # we only optimize for kernel hyperparameters
#             covar_x = self.base_covar(x)
#             mean_x = self._mean_module(x)
#         else:
#             covar_x = self.covar_module(x)
#             mean_x = self.mean_module(x)
#         return MultivariateNormal(mean_x, covar_x)


# class VariationalGaussianProcess(gpytorch.models.ApproximateGP):
#     """
#     A variational GP to compute the position of the inducing points.
#     We only optimize for the position of the continuous dimensions and keep the categorical dimensions constant.
#     """

#     def __init__(self, kernel: Kernel, X_inducing: torch.Tensor):
#         """
#         Initialize a Variational GP
#         we set the lower bound and upper bounds of inducing points for numerical hyperparameters between 0 and 1,
#         that is, we constrain the inducing points to lay inside the subregion.

#         Parameters
#         ----------
#         kernel: Kernel
#             kernel of the variational GP, its hyperparameter needs to be fixed when it is by LGPGA
#         X_inducing: torch.tensor (N_inducing, D)
#             inducing points
#         """
#         variational_distribution = gpytorch.variational.TrilNaturalVariationalDistribution(X_inducing.size(0))
#         variational_strategy = gpytorch.variational.VariationalStrategy(
#             self, X_inducing, variational_distribution, learn_inducing_locations=True
#         )
#         super(VariationalGaussianProcess, self).__init__(variational_strategy)
#         self.mean_module = gpytorch.means.ZeroMean()
#         self.covar_module = kernel

#         shape_X_inducing = X_inducing.shape
#         lower_X_inducing = torch.zeros([shape_X_inducing[-1]]).repeat(shape_X_inducing[0])
#         upper_X_inducing = torch.ones([shape_X_inducing[-1]]).repeat(shape_X_inducing[0])

#         self.variational_strategy.register_constraint(
#             param_name="inducing_points",
#             constraint=Interval(lower_X_inducing, upper_X_inducing, transform=None),
#         )
#         self.double()

#         for p_name, t in self.named_hyperparameters():
#             if p_name != "variational_strategy.inducing_points":
#                 t.requires_grad = False

#     def forward(self, x: torch.Tensor) -> MultivariateNormal:
#         """
#         Pass the posterior mean and variance given input X

#         Parameters
#         ----------
#         x: torch.Tensor
#             Input data
#         Returns
#         -------
#         """
#         mean_x = self.mean_module(x)
#         covar_x = self.covar_module(x, cont_only=True)
#         return MultivariateNormal(mean_x, covar_x)



from __future__ import annotations

import warnings

import numpy as np
from ConfigSpace import (
    Categorical,
    Configuration,
    ConfigurationSpace,
    EqualsCondition,
    Float,
    Integer,
)
from ConfigSpace.conditions import InCondition
from sklearn.model_selection import StratifiedKFold, cross_val_score
from sklearn.neural_network import MLPClassifier
from src.models.model import Model


class MLPModel(Model):
    @property
    def configspace(self) -> ConfigurationSpace:
        # Build Configuration Space which defines all parameters and their ranges.
        # To illustrate different parameter types, we use continuous, integer and categorical parameters.
        cs = ConfigurationSpace()

        n_layer = Integer("n_layer", (1, 5), default=1)
        n_neurons = Integer("n_neurons", (8, 256), log=True, default=10)
        activation = Categorical("activation", ["logistic", "tanh", "relu"], default="tanh")
        solver = Categorical("solver", ["lbfgs", "sgd", "adam"], default="adam")
        batch_size = Integer("batch_size", (30, 300), default=200)
        learning_rate = Categorical("learning_rate", ["constant", "invscaling", "adaptive"], default="constant")
        learning_rate_init = Float("learning_rate_init", (0.0001, 1.0), default=0.001, log=True)

        # Add all hyperparameters at once:
        cs.add([n_layer, n_neurons, activation, solver, batch_size, learning_rate, learning_rate_init])

        # Adding conditions to restrict the hyperparameter space...
        # ... since learning rate is used when solver is 'sgd'.
        use_lr = EqualsCondition(child=learning_rate, parent=solver, value="sgd")
        # ... since learning rate initialization will only be accounted for when using 'sgd' or 'adam'.
        use_lr_init = InCondition(child=learning_rate_init, parent=solver, values=["sgd", "adam"])
        # ... since batch size will not be considered when optimizer is 'lbfgs'.
        use_batch_size = InCondition(child=batch_size, parent=solver, values=["sgd", "adam"])

        # We can also add multiple conditions on hyperparameters at once:
        cs.add([use_lr, use_batch_size, use_lr_init])

        return cs

    def train(self, config: Configuration, seed: int = 0, budget: int = 25) -> float:
        assert self.dataset is not None
        # For deactivated parameters (by virtue of the conditions),
        # the configuration stores None-values.
        # This is not accepted by the MLP, so we replace them with placeholder values.
        lr = config["learning_rate"] if config["learning_rate"] else "constant"
        lr_init = config["learning_rate_init"] if config["learning_rate_init"] else 0.001
        batch_size = config["batch_size"] if config["batch_size"] else 200

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")

            classifier = MLPClassifier(
                hidden_layer_sizes=[config["n_neurons"]] * config["n_layer"],
                solver=config["solver"],
                batch_size=batch_size,
                activation=config["activation"],
                learning_rate=lr,
                learning_rate_init=lr_init,
                max_iter=int(np.ceil(budget)),
                random_state=seed,
            )

            # Returns the 5-fold cross validation accuracy
            cv = StratifiedKFold(n_splits=5, random_state=seed, shuffle=True)  # to make CV splits consistent
            score = cross_val_score(classifier, self.dataset.get_X(), self.dataset.get_Y(), cv=cv, error_score="raise")

        return 1 - np.mean(score)



from smac.random_design.abstract_random_design import AbstractRandomDesign
from smac.random_design.annealing_design import CosineAnnealingRandomDesign
from smac.random_design.modulus_design import (
    DynamicModulusRandomDesign,
    ModulusRandomDesign,
)
from smac.random_design.probability_design import (
    DynamicProbabilityRandomDesign,
    ProbabilityRandomDesign,
)

__all__ = [
    "AbstractRandomDesign",
    "CosineAnnealingRandomDesign",
    "ModulusRandomDesign",
    "DynamicModulusRandomDesign",
    "ProbabilityRandomDesign",
    "DynamicProbabilityRandomDesign",
]



class TargetAlgorithmAbortException(Exception):
    """Exception indicating that the target function suggests an ABORT of SMAC, usually because it
    assumes that all further runs will surely fail.
    """

    pass


class FirstRunCrashedException(TargetAlgorithmAbortException):
    """Exception indicating that the first run crashed (depending on options this could trigger an
    ABORT of SMAC).
    """

    pass



from __future__ import annotations

import numpy as np

from smac.model.abstract_model import AbstractModel
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class RandomModel(AbstractModel):
    """AbstractModel which returns random values on a call to `fit`."""

    def _train(self, X: np.ndarray, Y: np.ndarray) -> RandomModel:
        if not isinstance(X, np.ndarray):
            raise NotImplementedError("X has to be of type np.ndarray.")
        if not isinstance(Y, np.ndarray):
            raise NotImplementedError("Y has to be of type np.ndarray.")

        logger.debug("(Pseudo) fit model to data.")
        return self

    def _predict(
        self,
        X: np.ndarray,
        covariance_type: str | None = "diagonal",
    ) -> tuple[np.ndarray, np.ndarray | None]:
        if covariance_type != "diagonal":
            raise ValueError("`covariance_type` can only take `diagonal` for this model.")

        if not isinstance(X, np.ndarray):
            raise NotImplementedError("X has to be of type np.ndarray.")

        return self._rng.rand(len(X), 1), self._rng.rand(len(X), 1)



from smac.model.abstract_model import AbstractModel
from smac.model.multi_objective_model import MultiObjectiveModel
from smac.model.random_model import RandomModel

__all__ = [
    "AbstractModel",
    "MultiObjectiveModel",
    "RandomModel",
]



from __future__ import annotations

from typing import Any

import copy

import numpy as np

from smac.acquisition.function.abstract_acquisition_function import (
    AbstractAcquisitionFunction,
)
from smac.model.abstract_model import AbstractModel
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class IntegratedAcquisitionFunction(AbstractAcquisitionFunction):
    r"""Compute the integrated acquisition function by marginalizing over model hyperparameters

    See "Practical Bayesian Optimization of Machine Learning Algorithms" by Jasper Snoek et al.
    (https://papers.nips.cc/paper/4522-practical-bayesian-optimization-of-machine-learning-algorithms.pdf)
    for further details.

    Parameters
    ----------
    acquisition_function : AbstractAcquisitionFunction
        Acquisition function to be integrated.

    Attributes
    ----------
    _acquisition_function : AbstractAcquisitionFunction
        Acquisition function to be integrated.
    _functions: list[AbstractAcquisitionFunction]
        Holds n (n = number of models) copies of the acquisition function.
    _eta : float
        Current incumbent function value.

    """

    def __init__(self, acquisition_function: AbstractAcquisitionFunction) -> None:
        super().__init__()
        self._acquisition_function: AbstractAcquisitionFunction = acquisition_function
        self._functions: list[AbstractAcquisitionFunction] = []
        self._eta: float | None = None

    @property
    def name(self) -> str:  # noqa: D102
        return f"Integrated Acquisition Function ({self._acquisition_function.__class__.__name__})"

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update({"acquisition_function": self._acquisition_function.meta})

        return meta

    def _update(self, **kwargs: Any) -> None:
        """Update the acquisition functions values.

        This method will be called if the model is updated. For example, entropy search uses it to update its
        approximation of P(x=x_min) and EI uses it to update the current fmin.

        This implementation creates an acquisition function object for each model to integrate over and sets the
        respective attributes for each acquisition function object.

        Parameters
        ----------
        kwargs : Any
            Keyword arguments for the model.

        Raises
        ------
        ValueError
            If the number of models is zero.
        """
        model = self.model
        models: list[AbstractModel] | None = None
        if hasattr(model, "models"):
            models = model.models  # type: ignore

        if models is None or len(models) == 0:
            raise ValueError("IntegratedAcquisitionFunction requires at least one model to integrate!")

        if len(self._functions) == 0 or len(self._functions) != len(models):
            self._functions = [copy.deepcopy(self._acquisition_function) for _ in models]

        for submodel, func in zip(models, self._functions):
            func.update(model=submodel, **kwargs)

    def _compute(self, X: np.ndarray) -> np.ndarray:
        """Compute integrated acquisition values

        Parameters
        ----------
        X : np.ndarray [N, D]
            The input points where the acquisition function should be evaluated. The dimensionality of X is (N, D),
            with N as the number of points to evaluate at and D is the number of dimensions of one X.

        Returns
        -------
        np.ndarray [N,1]
            Acquisition function values wrt X.

        Raises
        ------
        ValueError
            If `update` has not been called first (`_functions` not up to date).
        """
        if self._functions is None:
            raise ValueError("Need to call `update` first!")

        return np.array([func._compute(X) for func in self._functions]).mean(axis=0)



from __future__ import annotations

from ConfigSpace import Configuration

from smac.acquisition.maximizer.abstract_acquisition_maximizer import (
    AbstractAcquisitionMaximizer,
)
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class RandomSearch(AbstractAcquisitionMaximizer):
    """Get candidate solutions via random sampling of configurations."""

    def _maximize(
        self,
        previous_configs: list[Configuration],
        n_points: int,
        _sorted: bool = False,
    ) -> list[tuple[float, Configuration]]:
        """Maximize acquisition function with random search

        Parameters
        ----------
        previous_configs : list[Configuration]
            Not used.
        n_points : int
            Number of configurations to return.
        _sorted : bool, optional
            If True, sort candidates by their acquisition value (descending), by default False

        Returns
        -------
        list[tuple[float, Configuration]]
            Candidates with their acquisition function value. (acq value, candidate)
        """
        if n_points > 1:
            rand_configs = self._configspace.sample_configuration(size=n_points)
        else:
            rand_configs = [self._configspace.sample_configuration()]

        if _sorted:
            for i in range(len(rand_configs)):
                rand_configs[i].origin = "Acquisition Function Maximizer: Random Search (sorted)"

            return self._sort_by_acquisition_value(rand_configs)
        else:
            for i in range(len(rand_configs)):
                rand_configs[i].origin = "Acquisition Function Maximizer: Random Search"

            return [(0, rand_configs[i]) for i in range(len(rand_configs))]
